#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import sys
import numpy as np
import warnings

from ..utility import UnimplementedInstance, find_row
from xcsi.csi._material import find_material
from shps.rotor import exp



def _is_truss(frame, csi):
    if "FRAME RELEASE ASSIGNMENTS 1 - GENERAL" in csi:
        release = find_row(csi["FRAME RELEASE ASSIGNMENTS 1 - GENERAL"],
                           Frame=frame["Frame"])
    else:
        return False

    return release and all(release[i] for i in ("TI", "M2I", "M3I", "M2J", "M3J"))



def add_frames(instance):
    csi = instance._tree # Parsed input file
    names = instance.names # class mapping from CSI names to Xara tags
    model = instance.model # the xara.Model
    asm = instance.assembly

    tags = {}
    transforms = _FrameAxes(instance)

    for frame in csi.get("CONNECTIVITY - FRAME",[]):

        if _is_truss(frame, csi):
            instance._log(UnimplementedInstance("Truss", frame))
            continue

        if "IsCurved" in frame and frame["IsCurved"]:
            instance._log(UnimplementedInstance("Frame.Curve", frame))

        


        automesh = find_row(
            csi.get("FRAME AUTO MESH ASSIGNMENTS", []),
            Frame=frame["Frame"]
        )

        #
        # 1) Nodes
        #
        springs = find_row(csi.get("FRAME RELEASE ASSIGNMENTS 2 - PARTIAL FIXITY", []), Frame=frame["Frame"])
        if springs:
            instance._log(
                UnimplementedInstance(f"Frame with springs", 
                                      frame,
                                      table="FRAME RELEASE ASSIGNMENTS 2 - PARTIAL FIXITY"))

        if False: #springs:
            inode = names.define("Joint", "node")
            jnode = names.define("Joint", "node")
            model.node(inode, *asm.create_node(
                find_row(
                    csi["JOINT COORDINATES"],
                    Joint=frame["JointI"]
                )
            ).location)
            model.node(jnode, *asm.create_node(
                find_row(
                    csi["JOINT COORDINATES"],
                    Joint=frame["JointJ"]
                )
            ).location)
            for i,dof in enumerate(["_", "V2", "V3", "_", "M2", "M3"]):
                for n in "IJ":
                    key = f"{dof}{n}"
                    if key in springs and springs[key] > 0.0:
                        mat = names.define("Material", "material")
                        model.uniaxialMaterial("Elastic", mat, springs[key])
                        model.spring(
                            inode if n == "I" else jnode,
                            i,
                            k=springs[key]
                        )
            nodes = (inode, jnode)
        else:
            nodes = (
                names.identify("Joint", "node", frame["JointI"]),
                names.identify("Joint", "node", frame["JointJ"])
            )

        xi = np.array(model.nodeCoord(nodes[0]))
        xj = np.array(model.nodeCoord(nodes[1]))
        for nodes in _mesh_nodes((xi, xj), nodes, automesh, instance):

            tags[frame["Frame"]] = create_frame(frame, nodes,
                                                instance=instance,
                                                transforms=transforms)

    # TODO: remove this
    names._library["frame_tags"] = tags

    return []


def _mesh_nodes(x, nodes, automesh, instance):
    """ 
    return a list of node pairs defining the meshed frame elements.

    TODO: Get tolerance from, eg,
    
TABLE:  "PREFERENCES - DIMENSIONAL"
   MergeTol=0.008333333333333333   FineGrid=1   Nudge=1   SelectTol=3   SnapTol=12   SLineThick=2   PLineThick=4   MaxFont=8   MinFont=6   AutoZoom=10   ShrinkFact=70   TextFileLen=240
 
    """
    model = instance.model
    names = instance.names
    csi = instance._tree
    xi, xj = x
    ni, nj = nodes
    u = xj - xi
    Lu = float(np.linalg.norm(u))
    if automesh is None:
        return [nodes]
    if not any(value for key, value in automesh.items() if key != "Frame"):
        return [nodes]
    if automesh.get("AtJoints", False):
        # Find all joints (nodes) lying on the segment [xi, xj]
        intersecting = []

        for tag in model.getNodeTags():
            if tag in nodes:
                continue

            xn = np.array(model.nodeCoord(tag), dtype=float)

            # Distance from point to line
            d_line = np.linalg.norm(np.cross(u, xn - xi)) / Lu
            if d_line >= 1e-6:
                continue

            # Ensure it's within the segment (projection parameter t in [0,1])
            t = float(np.dot(xn - xi, u) / (Lu * Lu))
            if t <= 0.0 or t >= 1.0:
                continue

            # Also ensure point is actually close to the segment endpoints numerically
            # (optional; projection test is usually enough)
            intersecting.append((t, tag))

        # Sort by position along the element
        intersecting.sort(key=lambda p: p[0])

        # Build the ordered node list: i, internal..., j
        chain = [ni] + [tag for _, tag in intersecting] + [nj]

        # Convert into consecutive pairs
        pairs = [(chain[k], chain[k + 1]) for k in range(len(chain) - 1)]

        # (Optional) sanity: remove accidental duplicates (same tag repeated)
        pairs = [(a, b) for a, b in pairs if a != b]
        return pairs
    else:
        instance._log(UnimplementedInstance("Frame.AutoMesh.NonJoint", automesh))
    return [nodes]



def create_frame(frame, nodes, instance, transforms):
    csi = instance._tree # Parsed input file
    names = instance.names # class mapping from CSI names to Xara tags
    model = instance.model # the xara.Model
    asm = instance.assembly

    frame_obj = asm.create_frame(frame)

    options = instance.options.find(Frame=frame["Frame"])

    #
    # 1) Nodes
    #
    springs = find_row(csi.get("FRAME RELEASE ASSIGNMENTS 2 - PARTIAL FIXITY", []), Frame=frame["Frame"])
    if springs:
        instance._log(
            UnimplementedInstance(f"Frame with springs", 
                                    frame,
                                    table="FRAME RELEASE ASSIGNMENTS 2 - PARTIAL FIXITY"))

    if False: #springs:
        inode = names.define("Joint", "node")
        jnode = names.define("Joint", "node")
        model.node(inode, *asm.create_node(
            find_row(
                csi["JOINT COORDINATES"],
                Joint=frame["JointI"]
            )
        ).location)
        model.node(jnode, *asm.create_node(
            find_row(
                csi["JOINT COORDINATES"],
                Joint=frame["JointJ"]
            )
        ).location)
        for i,dof in enumerate(["_", "V2", "V3", "_", "M2", "M3"]):
            for n in "IJ":
                key = f"{dof}{n}"
                if key in springs and springs[key] > 0.0:
                    mat = names.define("Material", "material")
                    model.uniaxialMaterial("Elastic", mat, springs[key])
                    model.spring(
                        inode if n == "I" else jnode,
                        i,
                        k=springs[key]
                    )
        nodes = (inode, jnode)

    xi = np.array(model.nodeCoord(nodes[0]))
    xj = np.array(model.nodeCoord(nodes[1]))

    if np.linalg.norm(xj - xi) < 1e-10:
        instance._log(UnimplementedInstance("Frame.ZeroLength", frame))
        return None
    #
    # 2) Section
    #
    sect_asgn = find_row(csi["FRAME SECTION ASSIGNMENTS"],
                            Frame=frame["Frame"])
    
    section = frame_obj.section

    #
    # 3) Geometric Transformation
    #
    transform: int = transforms.find(frame, xi, xj)

    #
    # 4) Mass
    #

    total_mass = frame_obj.mass / frame_obj.length


    #
    # 5) Create the element
    #

    releases = find_row(csi.get("FRAME RELEASE ASSIGNMENTS 1 - GENERAL", []), Frame=frame["Frame"])
    if releases and any(releases[i] for i in ("TI", "M2I", "M3I", "TJ", "M2J", "M3J")):
        names.log(UnimplementedInstance("Frame with releases", 
                                        frame,
                                        table="FRAME RELEASE ASSIGNMENTS 1 - GENERAL"))


    # Case 1: Inelastic frame
    if instance is not None and options.inelastic:

        if section.has_multiaxial:
            stag = names.identify("AnalSect", "section", f"Multiaxial/{section.name}")
            tag = names.define("Frame", "element", frame["Frame"])

            model.element("ForceFrame",
                            tag,
                            nodes,
                            transform=transform,
                            section=stag,
                            mass=total_mass,
                            shear=int(options.use_shear)
            )
            return tag


    # Case 2: Elastic Prismatic

    # TODO: probably dont use "SectionType" from this table; it looks like its 
    # superceded by "Shape" in Section Properties 01 table.
    elif frame_obj.prismatic:

        section = names.identify("AnalSect", "section", sect_asgn["AnalSect"])
        if section is None:
            warnings.warn(f"No section found for {sect_asgn['AnalSect']}")
            return None
        
        tag = names.define("Frame", "element", frame["Frame"])

        model.element("ForceFrame",
                        tag,
                        nodes,
                        section=section,
                        # n=8,
                        # gauss_type="Legendre",
                        transform=transform,
                        mass=total_mass,
                        shear=int(options.use_shear),
                        comment=f"Frame {frame['Frame']}"
        )
        return tag

    # Case 3: Elastic Non-prismatic

    elif sect_asgn["NPSectType"] == "Default" and (
            integr := names.identify("AnalSect", "integration", sect_asgn["AnalSect"])):
        # Non-prismatic sections
        tag = names.define("Frame", "element", frame["Frame"])
        model.element("ForceFrame",
                        tag,
                        nodes,
                        transform,
                        integr,
                        mass=total_mass,
                        shear=int(options.use_shear)
        )
        return tag

    else:
        names.log(UnimplementedInstance("FrameSection.NPSectType", sect_asgn["NPSectType"]))
        return None



def _find_frame_mass(csi, frame, sect_info, total_length):

    is_nonprismatic = (sect_info["Shape"] == "Nonprismatic")

    if "FRAME ADDED MASS ASSIGNMENTS" in csi:
        row = find_row(csi["FRAME ADDED MASS ASSIGNMENTS"],
                        Frame=frame["Frame"])
        additional_mass = row["MassPerLen"] if row else 0.0
    else:
        additional_mass = 0.0

    if not is_nonprismatic:
        #
        # Prismatic section
        #
        A = sect_info["Area"]  # cross‐sectional area
        mat_info = find_material(csi, sect_info["Material"])
        rho = mat_info["density"]  # material density (mass / volume)

        # self‐weight mass per length = area × density
        self_weight_mpl = A * rho

    else:
        #
        # Nonprismatic section
        #
        # look for total mass and total length in
        #

        # total mass for the entire nonprismatic section
        total_mass = sect_info.get("TotalMass", 0)

        # self‐weight mass per length = total mass / total length
        self_weight_mpl = total_mass / total_length

    # Combine any assigned mass per length with self‐weight mass per length
    total_mass = self_weight_mpl + additional_mass

    return {
        "total_mass": total_mass,
        "mass_type": "lumped"
    }



def _orient(xi, xj, angle):
    """
    Calculate the coordinate transformation vector.
    xi is the location of node I, xj node J,
    and `angle` is the rotation about the local axis in degrees

    By default local axis 2 is always in the 1-Z plane, except if the object
    is vertical and then it is parallel to the global X axis.
    The definition of the local axes follows the right-hand rule.
    """

    # The local 1 axis points from node I to node J
    e1 = xj - xi
    dx, dy, dz = e1
    e1 = e1 / np.linalg.norm(e1)

    # Global z
    E3 = np.array([0, 0, 1])

    # In Sap2000, if the element is vertical, the local y-axis is the same as the
    # global x-axis
    if dx == 0 and dy == 0:
        e2 = np.array([1, 0, 0])
        e3 = np.cross(e1, e2)

    # Otherwise, the plane composed of the local x-axis and the local
    # y-axis is a vertical plane. In this
    # case, the local z-axis is the cross product of the local
    # x-axis and the global z-axis.
    else:
        e3 = -np.cross(E3, e1)
        e2 = np.cross(e3, e1)

    e3 = e3 / np.linalg.norm(e3)
    e2 = e2 / np.linalg.norm(e2)


    # Rotate the local axis using the Rodrigue rotation formula
    # convert from degrees to radians
    angle = angle / 180 * np.pi
    # return exp(angle * e1 / np.linalg.norm(e1)) @ e3
    e3r = e3 * np.cos(angle) + np.cross(e1, e3) * np.sin(angle)
    # Finally, the normalized local z-axis is returned
    return e3r / np.linalg.norm(e3r)


class _FrameAxes:
    def __init__(self, instance):
        self._ins = instance
        self._ndm = instance.assembly.ndm
        self._transforms = {}
        self._model = instance.model
        self._ttag = 1
    
    def find(self, frame, xi, xj):
        if self._ndm != 3:
            vecxz = [0, 0, 1]
        else:
            vecxz = self._orient(frame, xi, xj)

        if tuple(vecxz) not in self._transforms:
            # TODO: Use frame options to set linear/pdelta/corotational 
            self._model.geomTransf("Linear", self._ttag, *vecxz)
            self._transforms[tuple(vecxz)] = self._ttag
            self._ttag += 1
        return self._transforms[tuple(vecxz)]
    

    def _orient(self, frame, xi, xj):
        csi = self._ins._tree
        if "FRAME LOCAL AXES ASSIGNMENTS 1 - TYPICAL" not in csi:
            return _orient(xi, xj, 0.0)

        row = find_row(csi["FRAME LOCAL AXES ASSIGNMENTS 1 - TYPICAL"],
                        Frame=frame["Frame"])

        # TODO: implement advanced axes and uncomment this line.
        if row and row.get("AdvanceAxes", False):
            # self._ins._log(UnimplementedInstance("Frame.LocalAxes.AdvancedAxes", frame))
            return self._orient_advanced(frame)

        angle = row["Angle"] if row else 0.0

        return _orient(xi, xj, angle-90)


    def _orient_advanced(self, frame):
        tables = self._ins._tree
        table = find_row(
            tables.get("FRAME LOCAL AXES ASSIGNMENTS 2 - ADVANCED", []),
            Frame=frame["Frame"]
        )
        if table is None:
            raise ValueError("Advanced frame axes requested but no ADVANCED table found.")
        
        R0 = self._ins.assembly._coordinate_system(table["PlCoordSys"]).rotation_matrix

        return R0[:,0]

    def _orient_angle(self, xi, xj, angle):
        """
        Calculate the coordinate transformation vector.
        xi is the location of node I, xj node J,
        and `angle` is the rotation about the local axis in degrees

        By default local axis 2 is always in the 1-Z plane, except if the object
        is vertical and then it is parallel to the global X axis.
        The definition of the local axes follows the right-hand rule.
        """

        # The local 1 axis points from node I to node J
        e1 = xj - xi
        dx, dy, dz = e1
        e1 = e1 / np.linalg.norm(e1)

        # Global z
        E3 = np.array([0, 0, 1])

        # In Sap2000, if the element is vertical, the local y-axis is the same as the
        # global x-axis
        if dx == 0 and dy == 0:
            e2 = np.array([1, 0, 0])
            e3 = np.cross(e1, e2)

        # Otherwise, the plane composed of the local x-axis and the local
        # y-axis is a vertical plane. In this
        # case, the local z-axis is the cross product of the local
        # x-axis and the global z-axis.
        else:
            e3 = -np.cross(E3, e1)
            e2 = np.cross(e3, e1)

        e3 = e3 / np.linalg.norm(e3)
        e2 = e2 / np.linalg.norm(e2)


        # Rotate the local axis using the Rodrigue rotation formula
        # convert from degrees to radians
        angle = angle / 180 * np.pi
        # return exp(angle * e1 / np.linalg.norm(e1)) @ e3
        e3r = e3 * np.cos(angle) + np.cross(e1, e3) * np.sin(angle)
        # Finally, the normalized local z-axis is returned
        return e3r / np.linalg.norm(e3r)




class UnimplementedInstance:
    def __init__(self, name, object=None, table=None, issue=None):
        self.name = name 
        self.object = object
        self.table = table
        self.issue = issue

    def __repr__(self):
        return f"{self.name}: {self.object}"


def find_row(table, **kwds) -> dict:

    for row in table:
        match = True
        for k, v in kwds.items():
            if k not in row or row[k] != v:
                match = False
                break

        if match:
            return row


def find_rows(table, **kwds) -> list:
    rows = []
    for row in table:
        match = True
        for k, v in kwds.items():
            if k not in row or row[k] != v:
                match = False
                break

        if match:
            rows.append(row)

    return rows




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
class ModelOptions:
    def __init__(self, instance, options_list):
        self._options = {
            "Frame": []
        }
        self.__instance = instance
    
    def find(self, **kwds):
        if "Frame" in kwds:
            name = kwds["Frame"]
            return _FrameOptions(name, self.__instance._tree, self.__instance)


class ConfigNL01(ModelOptions):
    "Material Nonlinearity"
    def __init__(self, instance, options_list):
        super().__init__(instance, options_list)

class ConfigNL02(ModelOptions):
    "Geometric Nonlinearity"
    def __init__(self, instance, options_list):
        super().__init__(instance, options_list)

class ConfigNL03(ModelOptions):
    "Geometric and Material Nonlinearity"
    def __init__(self, instance, options_list):
        super().__init__(instance, options_list)

class _FrameOptions:
    use_shear: bool = True
    inelastic: str  = None  # "Uniaxial", "Multiaxial", "Hinge"
    prismatic: bool = True

    class Geometry:
        Linear       = 0 
        PDelta       = 1
        Corotational = 1
        Exact        = 2

    def __init__(self, name, csi, instance=None, options=None):
        if options is None:
            options = {}

        self.name = name
        self._csi = csi
        self._instance = instance

        self._geometry: _FrameOptions.Geometry = options.get("geometry", None)
        self._use_shear: bool = options.get("use_shear", None) 
        self._inelastic: bool = options.get("inelastic", None)
        self._element : str = options.get("element", None)

    @property
    def use_shear(self):
        # Timoshenko beam theory.

        if self.inelastic:
            return True

        from xcsi.csi import find_row
        csi = self._csi
        sect_asgn = find_row(csi["FRAME SECTION ASSIGNMENTS"],
                             Frame=self.name)
        section = self._instance.find(frame_section=sect_asgn["AnalSect"])
        if section is None:
            return True
        return section.has_shear

    @property 
    def geometry(self):
        if self._geometry is not None:
            return self._geometry
        return _FrameOptions.Geometry.Linear


    @property
    def inelastic(self)->bool:
        "FRAME HINGE ASSIGNS 01 - OVERVIEW"
        "FRAME HINGE ASSIGNS 00 - HINGE DISTRIBUTION TYPE"
        return False
        hinge = find_row(self._csi.get("FRAME HINGE ASSIGNS 00 - HINGE DISTRIBUTION TYPE",[]), 
                         Frame=self.name)

        if hinge and "Nonlinear" in hinge.get("HgeDistrTyp",""):
            return True
        
        hinge = find_row(self._csi.get("FRAME HINGE ASSIGNS 01 - OVERVIEW",[]), 
                         Frame=self.name)
        if hinge and "Nonlinear" in hinge.get("HgeDistrTyp",""):
            return True

        return False


#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import xara
import numpy as np
import warnings
from ..csi.utility import UnimplementedInstance, find_row, find_rows


class _LinkMaterial:
    _link_tables = {
    #   LinkType                                    TableName
        "Linear"               : "LINK PROPERTY DEFINITIONS 02 - LINEAR",
        "??                  " : "LINK PROPERTY DEFINITIONS 03 - MULTILINEAR",
        "Damper - Exponential" : "LINK PROPERTY DEFINITIONS 04 - DAMPER",
        "Gap"                  : "LINK PROPERTY DEFINITIONS 05 - GAP",
        "????                " : "LINK PROPERTY DEFINITIONS 06 - HOOK",
        "?????               " : "LINK PROPERTY DEFINITIONS 07 - RUBBER ISOLATOR",
        "??????              " : "LINK PROPERTY DEFINITIONS 08 - SLIDING ISOLATOR",
        "Plastic (Wen)"        : "LINK PROPERTY DEFINITIONS 10 - PLASTIC (WEN)",
        "MultiLinear Plastic"  : "LINK PROPERTY DEFINITIONS 11 - MULTILINEAR PLASTIC",
    }
    def __init__(self, prop_table, instance):
        self.name = prop_table["LinkProp"]
        self._gen_prop_table = prop_table
        self._dofs = []
        self._tags = []
        self._instance = instance
        self.create(instance)

    @property 
    def property_table(self):
        return self._link_tables.get(self._gen_prop_table["LinkType"], None)

    @property 
    def material_dofs(self):
        return tuple(["U1", "U2", "U3", "R1", "R2", "R3"].index(i)+1 for i in self._dofs)
    
    @property
    def material_tags(self):
        return tuple(self._tags)


class _LinkMaterial02(_LinkMaterial):
    def create(self, instance):
        csi = instance._tree
        model: xara.Model = instance.model
        names  = instance.names

        for dof in ["U1", "U2", "U3", "R1", "R2", "R3"]:
            table = find_row(csi["LINK PROPERTY DEFINITIONS 02 - LINEAR"],
                            Link=self.name, DOF=dof)
            if table is None:
                continue

            if dof in {"U1", "U2", "U3"}:
                stiff = table["TransKE"]
                damp  = table["TransCE"]
            elif dof in {"R1", "R2", "R3"}:
                stiff = table["RotKE"]
                damp  = table["RotCE"]
            
            if damp != 0:
                instance._log(UnimplementedInstance(f"LinkMaterial02.Damping", table))
            
            tag = names.define("LinkProp", "material", f"Link/{self.name}.{dof}")
            model.uniaxialMaterial(
                "Elastic",
                tag,
                stiff
            )
            self._dofs.append(dof)
            self._tags.append(tag)

        return

class _LinkMaterial05(_LinkMaterial):
    def create(self, instance):
        csi = instance._tree
        model  = instance.model
        names  = instance.names

        for dof in ["U1", "U2", "U3", "R1", "R2", "R3"]:
            table = find_row(csi["LINK PROPERTY DEFINITIONS 05 - GAP"],
                            Link=self.name, DOF=dof)
            if table is None:
                continue

            if dof in {"U1", "U2", "U3"}:
                stiff = table.get("TransKE", None)
                damp  = table.get("TransCE", None)
            
            if damp != 0:
                instance._log(UnimplementedInstance(f"LinkMaterial02.Damping", table))

            if not table["Fixed"] and stiff is not None:
                assert stiff is not None, f"Stiffness must be defined for non-fixed gap link {self.name} DOF {dof}"
                tag = names.define("LinkProp", "material", f"Link/{self.name}.{dof}")
                model.uniaxialMaterial(
                    "Elastic",
                    tag,
                    stiff
                )
            else:
                tag = None

            self._dofs.append(dof)
            self._tags.append(tag)

        return

# _LinkMaterial04 - Damper (Exponential)
    # for damper in csi.get("LINK PROPERTY DEFINITIONS 04 - DAMPER", []):
    #     # TODO: implement dampers
    #     conv.log(UnimplementedInstance("Link.Damper", damper))
    #     continue
    #     name = damper["Link"]
    #     stiff = damper["TransK"]
    #     dampcoeff = damper["TransC"]
    #     exp = damper["CExp"]
    #     model.eval(f"uniaxialMaterial ViscousDamper {mat_total} {stiff} {dampcoeff}' {exp}\n")

    #     dof = damper["DOF"]
    #     library["link_materials"][name][dof] = mat_total
    #     mat_total += 1

class _LinkMaterial10(_LinkMaterial):
    def create(self, instance):
        csi = instance._tree
        model = instance.model
        names  = instance.names

        for dof in ["U1", "U2", "U3", "R1", "R2", "R3"]:

            table = find_row(csi["LINK PROPERTY DEFINITIONS 10 - PLASTIC (WEN)"],
                             Link=self.name, DOF=dof)

            # Skip if property not defined for this DOF
            if table is None:
                continue

            fy = table["YieldForce"]
            ratio = table["ratio"]
            stiff = table["TransK"]
            tag = names.define("LinkProp", "material", f"Link/{self.name}.{dof}")
            # TODO: use the actual Bouc-Wen model here
            model.uniaxialMaterial(
                "Steel01",
                tag,
                fy,
                stiff,
                ratio
            )
            self._dofs.append(dof)
            self._tags.append(tag)
        return
    
class _LinkMaterial11(_LinkMaterial):
    def create(self, instance):
        csi = instance._tree
        model = instance.model
        names  = instance.names
        for dof in ["U1", "U2", "U3", "R1", "R2", "R3"]:

            # tables = find_rows(csi["LINK PROPERTY DEFINITIONS 11 - MULTILINEAR PLASTIC"],
            #                   Link=self.name, DOF=dof)
            
            tables = find_row(csi["LINK PROPERTY DEFINITIONS 11 - MULTILINEAR PLASTIC"],
                              Link=self.name, DOF=dof, Point=1)
            if tables is None:
                continue
            
            K = tables["TransKE"]
            tag = names.define("LinkProp", "material", f"Link/{self.name}.{dof}")
            model.uniaxialMaterial(
                "Elastic",
                tag,
                K
            )
            self._dofs.append(dof)
            self._tags.append(tag)
        return
        points = []
        table = find_row(csi["LINK PROPERTY DEFINITIONS 11 - MULTILINEAR PLASTIC"],
                        Link=self.name, DOF=self.dof)
        
        if len(points) < 2:
            instance._log(UnimplementedInstance(f"LinkMaterial11.InsufficientPoints", table))
            return

        tag = names.define("LinkProp", "material", self.key)

        # https://opensees.berkeley.edu/wiki/index.php/MultiLinear_Material


class _LinkMaterialNA(_LinkMaterial):
    def create(self, instance):
        return


_link_tables = {
#   LinkType                                    TableName
    "Linear"               : _LinkMaterial02, # "LINK PROPERTY DEFINITIONS 02 - LINEAR",
    "??                  " : _LinkMaterialNA, # "LINK PROPERTY DEFINITIONS 03 - MULTILINEAR",
    "Damper - Exponential" : _LinkMaterialNA, # "LINK PROPERTY DEFINITIONS 04 - DAMPER",
    "Gap"                  : _LinkMaterial05, # "LINK PROPERTY DEFINITIONS 05 - GAP",
    "????                " : _LinkMaterialNA, # "LINK PROPERTY DEFINITIONS 06 - HOOK",
    "?????               " : _LinkMaterialNA, # "LINK PROPERTY DEFINITIONS 07 - RUBBER ISOLATOR",
    "??????              " : _LinkMaterialNA, # "LINK PROPERTY DEFINITIONS 08 - SLIDING ISOLATOR",
    "Plastic (Wen)"        : _LinkMaterial10, # "LINK PROPERTY DEFINITIONS 10 - PLASTIC (WEN)",
    "MultiLinear Plastic"  : _LinkMaterial11, # "LINK PROPERTY DEFINITIONS 11 - MULTILINEAR PLASTIC",
}


def create_links(instance):
    csi = instance._tree
    model: xara.Model = instance.model
    names  = instance.names

    link_props = _create_link_materials(instance)

    # Create Single Joint Links
    _create_single_links(instance, link_props)

    # Create Double Joint Links
    _create_double_links(instance, link_props)


def _orient(xi, xj, deg):
    # Calculate the direction vector of the link element
    # Where a is the node number of node i, b is the node number of node j, and degree is the user-specified local axis
    # ------------------------------------------------------------------------------
    d_x, d_y, d_z = xj - xi

    # Local 1-axis points from node I to node J
    l_x = np.array([d_x, d_y, d_z])
    # Global z-axis
    g_z = np.array([0, 0, 1])

    # In SAP2000, if the link is vertical, the local y-axis is the same as the
    # global x-axis, and the local z-axis can be obtained by crossing the local
    # x-axis with the local y-axis
    if d_x == 0 and d_y == 0:
        l_y = np.array([1, 0, 0])
        l_z = np.cross(l_x, l_y)

    # In other cases, the plane formed by the local x-axis and the local y-axis
    # is a vertical plane (i.e., the normal vector is horizontal), and the
    # local z-axis can be obtained by crossing the local x-axis with the global
    # z-axis
    else:
        l_z = np.cross(l_x, g_z)

    # The local axis may also be rotated using the Rodrigues' rotation formula
    angle = deg / 180 * np.pi
    l_z_rot = l_z * np.cos(angle) + np.cross(l_x, l_z) * np.sin(angle)
    # The rotated local y-axis can be obtained by crossing the rotated local z-axis with the local x-axis
    l_y_rot = np.cross(l_z_rot, l_x)
    # Finally, return the normalized local y-axis
    return l_y_rot / np.linalg.norm(l_y_rot)


def _create_link_materials(instance):
    csi = instance._tree

    links = {}

    for link in csi.get("LINK PROPERTY ASSIGNMENTS",[]):
        if link["LinkType"] not in _link_tables:
            instance._log(UnimplementedInstance(f"Link.Prop.LinkType={link['LinkType']}", link))
            continue
        
        if link["Link"] in links:
            continue

        cls = _link_tables[link["LinkType"]](link, instance)
        links[link["LinkProp"]] = cls
    return links


def _create_single_links(instance, link_props):

    csi = instance._tree
    model = instance.model
    names  = instance.names

    for link in csi.get("CONNECTIVITY - LINK",[]):

        assign = find_row(csi["LINK PROPERTY ASSIGNMENTS"],
                          Link=link["Link"])

        if assign.get("LinkJoints", "") != "SingleJoint":
            continue

        nodes = (
            names.identify("JointBase", "node", link["JointI"]),
            names.identify("JointBase", "node", link["JointJ"])
        )

        assert nodes[0] == nodes[1], f"SingleJoint link must connect the same node, got {link['JointI']} and {link['JointJ']} for {link['Link']}"

        # if any(not (isinstance(node, int) or node.isdigit()) for node in nodes):
        #     conv.log(UnimplementedInstance(f"Joint: non-integer nodes", assign))
        #     continue

        if assign["LinkProp"] not in link_props:
            instance._log(UnimplementedInstance(f"Link.Prop.Missing", assign))
            continue
        if isinstance(link_props[assign["LinkProp"]], _LinkMaterialNA):
            instance._log(UnimplementedInstance(f"Link.Prop.LinkType={assign['LinkType']}", assign))
            continue
        
        # Get instance of _LinkMaterial
        link_prop = link_props[assign["LinkProp"]]


        # TODO: Implement soil springs

        flags = [0 for _ in range(6)]
    
        for dof in link_prop.material_dofs:
            flags[dof-1] = 1

        model.fix(names.identify("JointBase", "node", link["JointI"]), tuple(flags))


def _create_double_links(instance, link_props):
    csi = instance._tree
    model: xara.Model = instance.model
    names  = instance.names

    for link in csi.get("CONNECTIVITY - LINK",[]):

        nodes = (
            names.identify("Joint", "node", link["JointI"]),
            names.identify("Joint", "node", link["JointJ"])
        )

        # if any(not (isinstance(node, int) or node.isdigit()) for node in nodes):
        #     conv.log(UnimplementedInstance(f"Joint: non-integer nodes", assign))
        #     continue

        assign = find_row(csi["LINK PROPERTY ASSIGNMENTS"],
                          Link=link["Link"])

        # Only process double-joint links here
        if assign.get("LinkJoints", "") != "TwoJoint":
            continue

        #
        # Get mats and dofs
        #
        if assign["LinkProp"] not in link_props:
            instance._log(UnimplementedInstance(f"Link.Prop.Missing", assign))
            continue
        if isinstance(link_props[assign["LinkProp"]], _LinkMaterialNA):
            instance._log(UnimplementedInstance(f"Link.Prop.LinkType={assign['LinkType']}", assign))
            continue

        link_prop = link_props[assign["LinkProp"]]

        # if the dof is fixed, then the tag will be None. we need to filter those

        mats = [
            mat for mat in link_prop.material_tags if mat is not None
        ]
        link_dofs = [
            dof for i,dof in enumerate(link_prop.material_dofs)
            if link_prop.material_tags[i] is not None
        ]
        fixed_dofs = [dof for i,dof in enumerate(link_prop.material_dofs)
            if link_prop.material_tags[i] is None
        ]
        for dof in fixed_dofs:
            model.fix(names.identify("JointBase", "node", link["JointI"]), dof=dof)
            model.fix(names.identify("JointBase", "node", link["JointJ"]), dof=dof)

        if len(link_dofs) == 0:
            instance._log(UnimplementedInstance(f"DoubleLink.DOFS", assign))
            continue
        if len(mats) != len(link_dofs):
            instance._log(UnimplementedInstance(f"DoubleLink.MatsVsDofs", assign))
            continue

        # Check whether there are any zero-length link elements
        xi = np.array(model.nodeCoord(nodes[0]))
        xj = np.array(model.nodeCoord(nodes[1]))

        distance = np.linalg.norm(xj - xi)
        zero_length_threshold = 1e-6

        #
        # Get axes and orientation
        #
        axes = find_row(csi.get("LINK LOCAL AXES ASSIGNMENTS 1 - TYPICAL", []),
                        Link=link["Link"])

        orient_vector = None  # Default value

        if axes:
            if axes["AdvanceAxes"]:
                # Handle advanced axes
                axes_advance = find_row(csi.get("LINK LOCAL AXES ASSIGNMENTS 2 - ADVANCED", []),
                                        Link=link["Link"])

                # Common advanced axes setup
                orient_vector = (
                    axes_advance["AxVecX"], axes_advance["AxVecY"], axes_advance["AxVecZ"],
                    axes_advance["PlVecX"], axes_advance["PlVecY"], axes_advance["PlVecZ"],
                )

                # Additional validation for non-zero-length links
                if distance > zero_length_threshold:
                    # Calculate node-based orientation
                    orient_vector_from_nodes = _orient(xi, xj, axes["Angle"])

                    # Calculate y-axis from advanced axes
                    ax_vec = np.array([axes_advance["AxVecX"], axes_advance["AxVecY"], axes_advance["AxVecZ"]])
                    pl_vec = np.array([axes_advance["PlVecX"], axes_advance["PlVecY"], axes_advance["PlVecZ"]])

                    x_axis = ax_vec / np.linalg.norm(ax_vec)
                    pl_projection = pl_vec - np.dot(pl_vec, x_axis) * x_axis
                    y_axis_advanced = pl_projection / np.linalg.norm(pl_projection)

                    # Validate and update orientation
                    if np.allclose(y_axis_advanced, orient_vector_from_nodes, atol=1e-6):
                        orient_vector = tuple(_orient(xi, xj, axes["Angle"]))
                    else:
                        orient_vector = tuple(_orient(xi, xj, axes["Angle"]))
                        warnings.warn(f"Orientation mismatch in link {link['Link']}")
#                       raise ValueError(f"Orientation mismatch in link {link['Link']}")

            else:
                # Handle typical axes (non-advanced)
                orient_vector = tuple(_orient(xi, xj, axes["Angle"]))


        #
        # Create the link
        #
        tag = names.define("Link", "element", link["Link"])
        if orient_vector is not None:
            model.element("TwoNodeLink",
                      tag,
                      nodes,
                      mat=mats,
                      dir=link_dofs,
                      orient=orient_vector
            )
        else:
            model.element("TwoNodeLink",
                      tag,
                      nodes,
                      mat=mats,
                      dir=link_dofs
            )

    return



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import xara
from xcsi.csi import create_model
from xcsi.names import Names
from .assembly import Assembly
from .options import ModelOptions

class _Units:
    pass


class _Instance:

    def __init__(self, job, 
                 model=None, 
                 options=None,
                 verbose=0):
        self._tree = job._tree
        self._job = job
        self._log_ = []
        self._names = Names()
        self._options = ModelOptions(self, options)
        self.assembly = job.assembly


        import xara
        if model is None:
            model = xara.Model(ndm=self.assembly.ndm, 
                               ndf=self.assembly.ndf)
        self._model = model

        create_model(job._tree,
                    model=self._model,
                    instance=self,
                    verbose=verbose)

    @property
    def units(self):
        # deprecated, access through job.assembly.units
        return self.assembly.units

    @property
    def names(self):
        return self._names

    @property
    def model(self)->xara.Model:
        return self._model
    
    @property 
    def options(self):
        return self._options


    def find(self, **kwds):
        return self._job.find(**kwds)
    
    def frame_outlines(self, frame_name):
        ...
    

    def _log(self, message):
        self._log_.append(message)
    
    def _print_log(self):
        import sys
        log = self._log_ + self._names._log

        types = {i.name for i in log}

        print("Unimplemented features", file=sys.stderr)
        for item in types:
            print(f"\t{item}: {sum(1 for i in log if i.name == item)}", file=sys.stderr)
        
        # for entry in log:
        #     print(f"\t{entry}", file=sys.stderr)

class Job:
    def __init__(self, tree):
        if not isinstance(tree, dict):
            self._filename = tree
            from xcsi._io import csiread
            tree = csiread(tree)
        self._tree = tree

    @property
    def units(self)->_Units:
    
        units = None
        unit_str = None
        for row in self._tree.get("PROGRAM CONTROL", []):
            if "CurrUnits" in row:
                unit_str = row["CurrUnits"]
        
        if unit_str.lower() == "kip, ft, f":
            import xara.units.fks as units
        elif unit_str.lower() == "kip, in, f":
            import xara.units.iks as units
        elif unit_str.lower() == "lb, in, f":
            import xara.units.ips as units
        elif unit_str.lower() == "kn, m, c":
            import xara.units.mks as units

        return units

    @property 
    def assembly(self):
        if not hasattr(self, "_assembly"):
            self._assembly = Assembly(self._tree)
        return self._assembly

    def find(self, **kwds):

        if "frame_section" in kwds:
            from xcsi.csi._frame.section import iter_sections
            for section,_ in iter_sections(self._tree):
                if section.name == kwds["frame_section"]:
                    return section
            
            for section,_ in iter_sections(self._tree, table="SECTION DESIGNER PROPERTIES 01 - GENERAL"):
                if section.name == kwds["frame_section"]:
                    return section

    def instance(self, model=None, options=None, verbose=0)->_Instance:
        return _Instance(self, model=model, options=options, verbose=verbose)


    def pattern(self, name):
        from xcsi._loading import LoadPattern
        return LoadPattern(name, self.assembly)

    def patterns(self):
        for pattern in self._tree.get("LOAD PATTERN DEFINITIONS", []):
            yield self.pattern(pattern["LoadPat"])

    def find_case(self, name):
        from xcsi.csi.analysis import (
            LinearStaticAnalysis, ModalAnalysis, NonlinearStaticAnalysis
        )
        for case in self._tree.get("LOAD CASE DEFINITIONS", []):
            if case["Case"] != name:
                continue
    
            if case["Type"] == "LinStatic":
                return LinearStaticAnalysis(case["Case"], self)

            elif case["Type"] == "LinDirHist":
                pass

            elif case["Type"] == "LinModal":
                return ModalAnalysis(case["Case"], self)

            elif case["Type"] == "NonStatic":
                return NonlinearStaticAnalysis(case["Case"], self)

            else:
                continue
                raise ValueError(f"Unsupported case type: {case['Type']}")

    def cases(self):
        from xcsi.csi.analysis import (
            LinearStaticAnalysis, ModalAnalysis, NonlinearStaticAnalysis
        )

        for case in self._tree.get("LOAD CASE DEFINITIONS", []):
            if not case.get("RunCase", False):
                continue

            if case["Type"] == "LinStatic":
                yield LinearStaticAnalysis(case["Case"], self)

            elif case["Type"] == "LinDirHist":
                pass

            elif case["Type"] == "LinModal":
                yield ModalAnalysis(case["Case"], self)

            elif case["Type"] == "NonStatic":
                yield NonlinearStaticAnalysis(case["Case"], self)

            else:
                continue
                raise ValueError(f"Unsupported case type: {case['Type']}")

    # Deprecated 
    steps = cases


#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import warnings
from .utility import find_row, find_rows
import numpy as np
from ._material import find_material


def _create_shell_integration(csi, model, conv):
    pass


def _create_shell_section(assign, instance):
    csi = instance._tree
    model = instance.model
    names = instance.names

    section = find_row(csi["AREA SECTION PROPERTIES"],
                        Section=assign["Section"]
    )

    tag = names.define("ShellSection", "section", assign["Section"])

    material = find_material(csi, section["Material"])

    model.section("ElasticShell", tag,
                    material["E"],  # E
                    material["v"],  # nu
                    section["Thickness"],
                    material["density"] if "density" in material else 0,
    )
    return tag


def create_shell_sections(instance):
    csi = instance._tree
    model = instance.model
    conv = instance.names


    for assign in csi.get("AREA SECTION ASSIGNMENTS", []):
        # if "Section" not in assign:
        #     print(assign)
        #     continue

        if not conv.identify("ShellSection", "section",  assign["Section"]):

            if not _create_shell_section(assign, instance):
                warnings.warn(f"Section {assign['Section']} not found")
                continue

            # tag += len(library["shell_sections"][assign["Section"]].integration)




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import sys
import json
from xcsi.csi import collect_outlines, find_rows
from xcsi.job import Job
from xcsi._io._json import DepthLimitedEncoder

class CommandLineError(Exception): pass

from xcsi._io import csiread



def main():
    read_file = None
    save_file = None
    operation = "View"
    strict_mode = None
    object_mode = "Model"
    interact = False
    Options = {
        "Execute": {
            "Case": None
        }
    }
    
    Verbose = 0
    show = set()
    objects = []
    fmt = "txt"

    i  = 0
    argi = iter(sys.argv[1:])
    for arg in argi:
        if arg == "-s":
            object_mode = "Section"
        elif arg == "-t":
            object_mode = next(argi)
            if object_mode not in ["Model", "Section", "Pattern", "Step"]:
                raise CommandLineError(f"Unknown object mode {object_mode}")

        elif arg == "-o":
            try:
                save_file = next(argi)
            except StopIteration:
                raise CommandLineError("Flag -o missing required argument <file>")

        elif arg == "-i":
            interact = True

        elif arg == "-m":
            try:
                strict_mode = next(argi)
            except StopIteration:
                raise CommandLineError("Flag -m missing required argument <mode>")

        elif arg == "--show":
            try:
                show.add(next(argi))
            except StopIteration:
                raise ValueError("Flag --show missing required argument <name>")
        
        elif arg == "--case":
            try:
                Options["Execute"]["Case"] = next(argi)
            except StopIteration:
                raise CommandLineError("Flag --case missing required argument <name>")
            try:
                Options["Execute"]["Case"] = int(Options["Execute"]["Case"])
            except ValueError:
                pass

        elif arg == "--verbose":
            try:
                Verbose = int(next(argi))
            except StopIteration:
                Verbose += 1
    
        elif arg == "--log":
            operation = "Log"
        elif arg[0] == "-":
            key = arg[1].lower()
            operation = {
                "p": "Print",
                "a": "Analyze",
                "c": "Export",
                "x": "Execute",
                "v": "Visualize",
                "r": "Visualize" # new; free up -v for verbose
            }[key]

        elif read_file is None:
            read_file = arg
        else:
            objects.append(arg.split("="))

        i += 1

    if read_file is None:
        raise ValueError("No file specified")

    csi = csiread(read_file)


    if operation == "Export":
        # Convert
        job = Job(csi)
        if object_mode == "Section":
            for k, v in objects:
                section = job.find(frame_section=v)
                if section is None:
                    raise ValueError(f"Section {v} not found")
                print(json.dumps(section._create_model().to_dict(), indent=2))
                sys.exit()


        if "tcl" in sys.argv[1]:
            import xara
            model = xara.Model(ndm=3, ndf=6, echo_file=sys.stdout)
            job.instance(model=model)

        elif "csi" in sys.argv[1]:
            if save_file is not None:
                with open(save_file, "w") as f:
                    json.dump(csi, f, cls=DepthLimitedEncoder, max_depth=2, indent=2)
            else:
                print(json.dumps(csi, cls=DepthLimitedEncoder, max_depth=2, indent=2))

        else:
            model = job.instance().model
            if save_file is not None:
                model.print("-json", f"{save_file}")
            else:
                import os
                if os.name == "nt":
                    stdout = "CON"
                else:
                    stdout = "/dev/stdout"
                model.print("-json", stdout)
        sys.exit()


    if operation == "Print":
        job = Job(csi)
        if object_mode == "Pattern":
            for pattern in job.patterns():
                print(pattern)
        elif object_mode == "Step":
            for step in job.steps():
                print(step)
        elif len(objects) == 0:
            for table in csi:
                print(f"{table}: {len(csi[table])}")
        else:
            for key, value in objects:
                for table in csi:
                    filt = {key: value}
                    if (row:= find_rows(csi[table], **filt)):
                        print(table)
                        for r in row:
                            for k, v in r.items():
                                if k == key:
                                    continue
                                print(f"  {k:<10}      {v}")
                            print()
                            continue
        sys.exit()



    if operation == "Execute":
        job = Job(csi)

        ins = job.instance(verbose=Verbose)

        if Options["Execute"]["Case"] is not None: 
            case = job.find_case(name=Options["Execute"]["Case"])
            if case is None:
                raise ValueError(f"Case {Options['Execute']['Case']} not found")
            case.run(ins)

        else:
            for step in job.steps():
                print(step)
                step.run(ins)
                ins.model.print()
            # break

        if interact:
            # Start an interactive console
            from opensees.repl.ptkshell import OpenSeesREPL
            repl = OpenSeesREPL(ins.model._openseespy._interp)
            repl.repl()

    elif operation == "Log":
        job = Job(csi)

        instance = job.instance(verbose=Verbose)
        instance._print_log()

    elif operation == "Visualize":
        job = Job(csi)

        # Visualize
        import veux
        instance = job.instance(verbose=Verbose)
        model = instance.model
        outlines = collect_outlines(csi, instance.names._library["frame_tags"])
        artist = veux.create_artist(model, canvas="gltf", vertical=3,
                    model_config={
                        "frame_outlines": outlines
                    }
        )
        # artist.draw_nodes()
        artist.draw_outlines()
        # artist.draw_surfaces()
        artist.draw_sections(outline=False)
        artist.draw_axes()
        artist.draw_origin(extrude=True)


        if save_file is not None:
            artist.save(save_file)
        else:
            veux.serve(artist)

    elif sys.argv[1] == "-Vn":
        # Visualize
        from scipy.linalg import null_space
        model.constraints("Transformation")
        model.analysis("Static")
        K = model.getTangent().T
        v = null_space(K)[:,0] #, rcond=1e-8)


        u = {
            tag: [1000*v[dof-1] for dof in model.nodeDOFs(tag)]
            for tag in model.getNodeTags()
        }

        import veux
        veux.serve(veux.render(model, u, canvas="gltf", vertical=3))

    elif sys.argv[1] == "-Q":
        # Quiet conversion
        pass
    else:
        raise ValueError(f"Unknown operation {sys.argv[1]}")


if __name__ == "__main__":
    main()



from xcsi.csi.utility import find_row
from shps.rotor import exp, log
import numpy as np

class Assembly:
    def __init__(self, csi):
        self._csi = csi 

        self.dofs = {key:val for key,val in csi["ACTIVE DEGREES OF FREEDOM"][0].items() }
        # self.dims = {key for key,val in csi["ACTIVE DEGREES OF FREEDOM"][0].items() }
        self.ndf = sum(1 for v in csi["ACTIVE DEGREES OF FREEDOM"][0].values())
        self.ndm = sum(1 
                       for k,v in csi["ACTIVE DEGREES OF FREEDOM"][0].items()
                        if k[0] == "U")
        
    @property 
    def tables(self):
        return self._csi

    @property 
    def units(self):
        units = None 
        unit_str = None
        for row in self._csi["PROGRAM CONTROL"]:
            if "CurrUnits" in row:
                unit_str = row["CurrUnits"]
        
        if unit_str.lower() == "kip, ft, f":
            import xara.units.fks as units
        elif unit_str.lower() == "kip, in, f":
            import xara.units.iks as units
        elif unit_str.lower() == "lb, in, f":
            import xara.units.ips as units
        elif unit_str.lower() == "kn, m, c":
            import xara.units.mks as units

        return units

    def _coordinate_system(self, name):
        return CooridnateSystem(
            find_row(self._csi.get("COORDINATE SYSTEMS", []),
                     Name=name)
        )

    def create_frame(self, frame_entry):
        """
        frame_entry: row in "CONNECTIVITY - FRAME" table
        """
        class Frame:
            def __init__(self, row, assembly: 'Assembly'):
                self.name = row["Frame"]
                self._row = row
                self._assembly = assembly
                self._csi = assembly.tables
                self._section: "FrameSection" = None
                self._section_assign = find_row(
                    self._csi["FRAME SECTION ASSIGNMENTS"],
                    Frame=self.name
                )
                self._section_general = find_row(
                        self._csi["FRAME SECTION PROPERTIES 01 - GENERAL"],
                        SectionName=self._section_assign["AnalSect"]
                )
                self._length = None

            @property 
            def prismatic(self)->bool:
                sect_asgn = self._section_assign
                return ("SectionType" not in sect_asgn)  or \
                       (sect_asgn["SectionType"] != "Nonprismatic") or \
                        sect_asgn["NPSectType"] == "Advanced"

            @property
            def mass(self)->float:
                """
                Return the total mass of the element.
                """
                csi = self._csi
                sect_info = self._section_general
                is_nonprismatic = (sect_info["Shape"] == "Nonprismatic")
                if self.section is None:
                    assert is_nonprismatic, "Section is None but section is prismatic"
                    # TODO
                    return 0.0

                self_weight_mpl = 0.0
                if "FRAME ADDED MASS ASSIGNMENTS" in csi:
                    row = find_row(csi["FRAME ADDED MASS ASSIGNMENTS"],
                                    Frame=self.name)
                    additional_mass = row["MassPerLen"] if row else 0.0
                else:
                    additional_mass = 0.0

                    # look for total mass and total length in
                    # total mass for the entire nonprismatic section

                    # self‐weight mass per length = total mass / total length
                    # self_weight_mpl += sect_info.get("TotalMass", 0) #/ self.length
                    # if not is_nonprismatic:
                        # self‐weight mass per length = area * density
                        # 1-014
                    self_weight_mpl += self.section.mass

                # Combine any assigned mass per length with self‐weight mass per length
                return self_weight_mpl*self.length + additional_mass*self.length


            @property
            def length(self)->float:
                if self._length is not None:
                    return self._length
                sect_asgn = self._section_assign
                if "NPSectLen" in sect_asgn:
                    total_length = sect_asgn["NPSectLen"]
                else:
                    # handle the case where NPSectLen doesn't exist
                    xi = self._assembly._make_node(
                        find_row(
                            self._csi["JOINT COORDINATES"],
                            Joint=self._row["JointI"]
                        )
                    ).location
                    xj = self._assembly._make_node(
                        find_row(
                            self._csi["JOINT COORDINATES"],
                            Joint=self._row["JointJ"]
                        )
                    ).location
                    total_length = np.linalg.norm(np.array(xj) - np.array(xi))
                self._length = total_length
                return self._length

            @property 
            def section(self):
                # This may return None if the element is non-prismatic
                if self._section is None:
                    sect_asgn = self._section_assign
                    self._section = self._assembly._frame_section(sect_asgn["AnalSect"])
                return self._section

        return Frame(frame_entry, self)

    def _frame_section(self, name):
        from xcsi.csi._frame.section import iter_sections
        for section,_ in iter_sections(self._csi):
            if section.name == name:
                return section
        
        for section,_ in iter_sections(self._csi, table="SECTION DESIGNER PROPERTIES 01 - GENERAL"):
            if section.name == name:
                return section


    def _make_node(self, coord_entry):
        """
        coord_entry: row in "JOINT COORDINATES" table
        """
        class Node:
            def __init__(self, row):
                self._row = row
            
            @property
            def location(self):
                if self._row["CoordSys"] != "GLOBAL":
                    return tuple(self._row[i] if i in self._row else 0.0 for i in ("GlobalX", "GlobalY", "GlobalZ"))
                else:
                    return tuple(self._row[i] if i in self._row else 0.0 for i in ("XorR", "Y", "Z"))

        return Node(coord_entry)



class CooridnateSystem:
    class Chord:
        @property 
        def plane(self):
            pass
        @property
        def angle(self):
            pass
        @property
        def axis(self):
            pass

    def __init__(self, row):
        self._row = row

    def chord(self, _):
        pass

    @property
    def rotation_vector(self):
        return log(self.rotation_matrix)
    
    @property
    def rotation_matrix(self):
        def _rad(deg):
            from math import pi
            return deg * pi / 180.0
        RZ = exp([0, 0, _rad(self._row["AboutZ"])])
        RY = exp([0, _rad(self._row["AboutY"]), 0])
        RX = exp([_rad(self._row["AboutX"]), 0, 0])
        return RZ @ RY @ RX




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
#
from .parse import load
import numpy as np
from .utility import UnimplementedInstance
from .._instance.link import create_links
from ._frame.outlines import collect_geometry as collect_outlines
from .utility import find_row, find_rows



def assemble(csi, model=None, verbose=False):
    from xcsi.job import Job
    return Job(csi).instance(verbose=verbose)



def create_model(csi,
                 model,
                 verbose=False,
                 instance=None):
    """
    Parameters
    ==========
    csi: a dictionary formed by calling ``csi.parse.load("file.b2k")``

    Returns
    =======
    model: xara.Model
    """
    # NOTE: This function was from an older design, and is being refactored
    # for the new Job/Instance oriented design.


    from .point import create_points
    from ._material import create_materials


    # Track used tables
    used = {
        "TABLES AUTOMATICALLY SAVED AFTER ANALYSIS"
    }

    #
    # Create model
    #

    used.add("ACTIVE DEGREES OF FREEDOM")

#   dofs = [f"U{i}" for i in range(1, ndm+1)]
#   if ndm == 3:
#       dofs = dofs + ["R1", "R2", "R3"]
#   else:
#       dofs = dofs + ["R3"]

    #
    # Create nodes
    #
    create_points(instance)

    #
    # Create materials and sections
    #
    create_materials(instance)

    _create_sections(instance)



    # Unimplemented objects
    for item in [
        "CONNECTIVITY - CABLE",
        "CONNECTIVITY - TENDON"]:
        for elem in csi.get(item, []):
            instance.names.log(UnimplementedInstance(item, elem))

    #
    # Add elements
    #
    _add_elements(instance)


    if verbose:
        instance._print_log()

    if verbose and False:
        for table in csi:
            if table not in used:
                print(f"\t{table}", file=sys.stderr)


    model.frame_tags = instance.names._library.get("frame_tags", {})

    return model


def _create_sections(instance):
    from ._section  import create_shell_sections
    from ._frame.section import add_frame_sections
    csi = instance._tree
    model = instance.model
    names = instance.names

    # 2) Frame
    add_frame_sections(csi, model, names)

    # 3) Shell
    create_shell_sections(instance)


def _add_elements(instance):
    """
    Add elements to the model.
    """
    from ._frame import add_frames
    from .._instance._shell import add_shells
    from ._solid import add_solids

    create_links(instance)
    add_frames(instance)
    add_shells(instance)
    add_solids(instance)



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
import json
import shlex

CONSTANTS = {
    "Yes": True,
    "No":  False
}

_SKIP = {
    "PROJECT INFORMATION",
    "FRAME DESIGN PROCEDURES",
    "PREFERENCES - STEEL DESIGN - AISC 360-10",
    "PREFERENCES - CONCRETE DESIGN - ACI 318-14",
    "PREFERENCES - CONCRETE SHELL DESIGN - EUROCODE 2-2004",
    "PREFERENCES - ALUMINUM DESIGN - AA-ASD 2000",
    "MATERIAL PROPERTIES 09 - ACCEPTANCE CRITERIA",
    "OPTIONS - COLORS - DISPLAY",
    "OPTIONS - COLORS - OUTPUT",
    "TABLES AUTOMATICALLY SAVED AFTER ANALYSIS",
}

def _parse_value(v):
    if v in CONSTANTS:
        return CONSTANTS[v]

    try:
        return json.loads(v)
    except:
        return v


def load(file, append: dict=None, skip=None, fmt: str=None):
    if fmt is None:
        fmt = "txt"

    if fmt == "xls":
        return _load_xls(file, append=append, skip=skip)

    return _load_txt(file, append=append, skip=skip)

def _load_xls(file, append: dict=None, skip=None)->dict:
    """
    Aug 15, 2025
    """
    from openpyxl import load_workbook
    wb = load_workbook(file, data_only=True, read_only=True)

    if append is None:
        tables = {}
    else:
        tables = append

    for ws in wb.worksheets:

        # Row 1: title (first non-empty cell)
        first_row = next(ws.iter_rows(min_row=1, max_row=1, values_only=True), ())
        if not first_row:
            continue
        title = next((str(v).strip() for v in first_row if v not in (None, "")), None)
        if not title:
            continue
        else:
            title = title.split(":")[1].strip().upper()

        # Row 2: headers
        header_row = next(ws.iter_rows(min_row=2, max_row=2, values_only=True), ())
        if not header_row:
            continue
        headers = [str(h).strip() if h not in (None, "") else None for h in header_row]
        # Take only non-empty headers
        col_idx = [i for i, h in enumerate(headers) if h]
        if not col_idx:
            continue
        header_names = [headers[i] for i in col_idx]

        # Rows 4+: data (skip row 3)
        records = []
        for r in ws.iter_rows(min_row=4, values_only=True):
            if r is None:
                continue

            vals = [(r[i] if i < len(r) else None) for i in col_idx]
            if all(v in (None, "") for v in vals):
                continue
            rec = {k: _parse_value(v) for k, v in zip(header_names, vals)}
            records.append(rec)

        if records:
            tables[title] = records

    return tables


def _load_txt(file, append: dict=None, skip:set=None)->dict:
    """
    Read file-like object file and form a dictionary. 

    Returns
    =======

    csi: A dictionary with the layout:
        {
           "TABLE NAME": [
            {"Key": "Val"},
            ...
           ]
        }
    """
    if skip is None:
        skip = _SKIP
    if append is None:
        tables = {}
    else:
        tables = append

    current_table = None
    current_item  = None
    for line_no, line in enumerate(file):
        if "END TABLE DATA" in line:
            break

        # Skip empty lines
        if line.isspace():
            continue

        if "TABLE:" in line:
            table_name = shlex.split(line)[1]
            current_item  = None

            # Append if table exists (append argument given)
            if table_name in tables:
                current_table = tables[table_name]
            else:
                current_table = []

            tables[table_name] = current_table


        # Data line
        elif current_table is not None:

            if current_item is None:

                current_item = {}
                current_table.append(current_item)

            continue_item = False
            try:
                # Things are complicated by the fact we have to parse things like:
                #    Key=Val'
                # into "Key": "Val'"
                #
                lex = shlex.shlex(line, posix=True)
                lex.quotes = '"'
                lex.wordchars += "'"
                lex.commenters = ""
                lex.whitespace_split = True
                tokens = list(lex)
            except Exception as e:
                raise ValueError(f"Error parsing line {line_no}: " + str(e))

            for i,kv in enumerate(tokens):
                # A "_" usually means a continuation of the previous line
                # but sometimes it is just a random character.
                if kv == "_":
                    if i == len(tokens)-1:
                        continue_item = True
                        break

                    # Sometimes there is a random "_" in the middle
                    # of a line?
                    else:
                        continue

                k, v = kv.split("=", maxsplit=1)
                current_item[k] = _parse_value(v)

            if not continue_item:
                current_item = None

    return tables




import xara
from xcsi.csi import find_row, find_rows
from collections import defaultdict
from dataclasses import dataclass

class _Pattern:
    @dataclass
    class JointLoad:
        joint: int
        force: list

    @dataclass
    class FrameLoad:
        pass

    def __init__(self):
        self.joint_forces = []
        self.frame_forces = []
        self.shell_forces = []


class LoadPattern:
    """
    "LOAD CASE DEFINITIONS"
    "LOAD PATTERN DEFINITIONS"

    "JOINT LOADS - FORCE"
    "FRAME LOADS - DISTRIBUTED" OK
    "FRAME LOADS - GRAVITY"
    "FRAME LOADS - POINT" TODO

    "CABLE LOADS - DISTRIBUTED"
    """
    def __init__(self, name, assembly):
        self.name = name
        csi = assembly.tables
        self._assembly = assembly

        joint_sp = defaultdict(dict)
        for joint in find_rows(csi.get("JOINT LOADS - GROUND DISPLACEMENT", []),
                               LoadPat=name):
            restraints = find_row(csi.get("JOINT RESTRAINT ASSIGNMENTS", []),
                                 Joint=joint["Joint"])
            for key in "U1", "U2", "U3", "R1", "R2", "R3":
                if key in joint:
                    if joint[key] == 0.0 and (restraints and not restraints[key]):
                        continue
                    joint_sp[joint["Joint"]][key] = joint[key]

        self._joint_sp = joint_sp

        self._read_joint_loads(csi)
        self._read_frame_loads(csi)


    def _read_joint_loads(self, csi):

        joint_force = defaultdict(lambda: [0.0 for _ in range(6)])
        for joint in find_rows(csi.get("JOINT LOADS - FORCE", []),
                               LoadPat=self.name):
            for i,key in enumerate(("F1", "F2", "F3", "M1", "M2", "M3")):
                if key in joint:
                    joint_force[joint["Joint"]][i] = joint[key]

        self._joint_force = joint_force


    def _apply_joint_loads(self, asm, ptag):
        model: xara.Model = asm.model
        names = asm.names

        for joint, force in self._joint_force.items():
            if not any(force):
                continue
            model.load(
                names.identify("JointBase", "node", joint),
                tuple(force),
                pattern=ptag
            )

    def _read_frame_loads(self, csi):
        name = self.name 

        # Frame loads
        self._frame_loads = defaultdict(list)
        for assign in find_rows(csi.get("FRAME LOADS - DISTRIBUTED", []),
                               LoadPat=name):
            
            if assign.get("RelDistA", 0) != 0:
                raise ValueError("Relative distance not supported for distributed frame loads")
            if assign.get("RelDistB", 1) != 1:
                raise ValueError("Relative distance not supported for distributed frame loads")
            if assign.get("DistType", "RelDist") != "RelDist":
                raise ValueError("Only relative distance supported for distributed frame loads")
            
            sign = 1.0
            direction = None
            try:
                direction = int(assign["Dir"])-1
            except ValueError:
                if assign["Dir"].lower() == "gravity":
                    direction = 2  # Z direction
                    sign = -1.0
                # elif assign["Dir"].lower() == "proj z":
                #     # Test 1-001, Case 4
                #     direction = 2  # Z direction
                #     frm = self._assembly.create_frame(
                #         find_row(
                #             csi.get("CONNECTIVITY - FRAME", []),
                #             Frame=assign["Frame"]
                #         )
                #     )
                #     sign = None
                else:
                    raise ValueError(f"Unsupported direction {assign['Dir']} for frame load {assign['Frame']}")

            force = [0,0,0]
            moment = [0,0,0]
            if assign.get("Type", None) == "Force":
                if assign["FOverLA"] != assign["FOverLB"]:
                    raise ValueError("FOverLA and FOverLB must be equal for distributed frame loads")

                force[direction] = sign * assign["FOverLA"]

            elif assign.get("Type", None) == "Moment":
                if assign["MOverLA"] != assign["MOverLB"]:
                    raise ValueError("MOverLA and MOverLB must be equal for distributed frame loads")

                raise NotImplementedError("Moment loads not implemented for frames")
            else:
                raise ValueError(f"Unsupported frame load type: {assign.get('Type', None)}")

            self._frame_loads[assign["Frame"]].append({
                "type": "Uniform",
                "basis": assign["CoordSys"].lower(),
                "force": force,
                "moment": moment,
                "offset": [0.0, 0.0, 0.0],
            })

        for assign in find_rows(csi.get("FRAME LOADS - GRAVITY", []),
                               LoadPat=name):
            sign = 1.0
            direction = None
            multipliers = [
                assign.get("MultiplierX", 0.0),
                assign.get("MultiplierY", 0.0),
                assign.get("MultiplierZ", 0.0),
            ]
            if sum(int(m!=0) for m in multipliers if m != 0.0) != 1.0:
                raise ValueError("Only one multiplier can be non-zero for distributed frame loads")
            direction = multipliers.index(next(m for m in multipliers if m != 0.0))
            sign = multipliers[direction]
            

            force = [0,0,0]
            moment = [0,0,0]
            frame = self._assembly.create_frame(
                find_row(
                    csi.get("CONNECTIVITY - FRAME", []),
                    Frame=assign["Frame"]
                )
            )
            mass = frame.mass
            gravity = self._assembly.units.gravity
            force[direction] = float(gravity*sign*mass/frame.length)

            self._frame_loads[assign["Frame"]].append({
                "type": "Uniform",
                "basis": assign["CoordSys"].lower(),
                "force": force,
                "moment": moment,
                "offset": [0.0, 0.0, 0.0],
            })

        for assign in find_rows(csi.get("FRAME LOADS - POINT", []),
                               LoadPat=name):
            if "RelDist" not in assign:
                raise ValueError("RelDist must be specified for point frame loads")

            xa = assign["RelDist"]
            sign = 1.0
            if assign["Dir"].lower() == "gravity":
                direction = 2  # Z direction
                sign = -1.0
            else:
                try:
                    direction = int(assign["Dir"])-1
                except ValueError:
                    raise ValueError(f"Unsupported direction {assign['Dir']} for frame load {assign['Frame']}")

            force = [0,0,0]
            moment = [0,0,0]
            if assign["Type"] == "Force":
                force[direction] = sign * assign["Force"]
            else:
                raise NotImplementedError("Moment loads not implemented for frames")

            self._frame_loads[assign["Frame"]].append({
                "type":  "Point",
                "basis":  assign["CoordSys"].lower(),
                "force":  force,
                "moment": moment,
                "offset": [xa, 0.0, 0.0],
            })

    def _apply_frame_loads(self, asm, ptag):
        model: xara.Model = asm.model
        names = asm.names
        for frame, loads in self._frame_loads.items():
            etag = names.identify("Frame", "element", frame)
            for load in loads:
                if load["type"] in {"Uniform", "Point"}:
                    model.eleLoad(
                        "Frame",
                        load["type"],
                        element=[etag],
                        offset=load["offset"],
                        basis=load["basis"],
                        force=load["force"],
                        couple=load["moment"],
                        pattern=ptag
                    )
                else:
                    raise NotImplementedError(f"Unsupported frame load type: {load['type']}")


    def apply(self, asm):
        """
        Apply the load pattern to the model.
        """
        model: xara.Model = asm.model
        dofs = ["U1", "U2", "U3", "R1", "R2", "R3"]

        ptag = asm.names.define("LoadPattern", "pattern", self.name)

        model.pattern("Plain", ptag, "Linear")
        for joint, sp in self._joint_sp.items():
            if not sp:
                continue
            for dof, value in sp.items():
                model.sp(
                    asm.names.identify("JointBase", "node", joint),
                    dofs.index(dof)+1,
                    value,
                    pattern=ptag
                )
        
        self._apply_frame_loads(asm, ptag)
        self._apply_joint_loads(asm, ptag)


    def __str__(self):
        return f"LoadPattern(name={self.name})"



class AmbientAcceleration:
    def __init__(self, name, csi):
        self.name = name
        self._csi = csi

        self.direction = None
        self.accel = None

        for row in find_rows(csi.get("LOAD PATTERN DEFINITIONS", []),
                             LoadPat=name):
            self.direction = row["Dir"]
            self.accel = row["Accel"]

        if self.direction is None or self.accel is None:
            raise ValueError(f"Ambient acceleration pattern {name} not found")

    def apply(self, asm):
        """
        Apply the ambient acceleration to the model.
        """
        model = asm.model
        dofs = ["U1", "U2", "U3"]

        ptag = asm.names.define("LoadPattern", "pattern", self.name)

        model.pattern("UniformExcitation", ptag, dofs.index(self.direction))
        model.ambient(self.direction, self.accel, pattern=ptag)

    def __str__(self):
        return f"AmbientAcceleration(name={self.name}, direction={self.direction}, accel={self.accel})"



from xcsi.csi import find_row, find_rows
from collections import defaultdict


class LinearStaticAnalysis:
    def __init__(self, name, job):
        self.name = name
        self._csi = job._tree
        self._job = job
        self.type = "LinStatic"


    def run(self, asm):
        """
        Run the linear static analysis case.
        """
        self._apply_loads(asm)

        # Run the analysis
        model = asm.model
        model.test("EnergyIncr", 1e-10, 10)
        model.system("Umfpack")
        model.constraints("Transformation")
        model.numberer("RCM")
        model.integrator("LoadControl", 1.0)
        model.algorithm("Newton")
        model.analysis("Static")
        model.analyze(1)
        model.reactions()
        model.print()

    def _apply_loads(self, asm):
        """
        Apply loads to the model for this linear static analysis case.
        """

        for assign in find_rows(self._csi.get("CASE - STATIC 1 - LOAD ASSIGNMENTS", []),
                       Case=self.name):
            pattern = self._job.pattern(assign["LoadName"])

            pattern.apply(asm)


    def __str__(self):
        return f"LinearStaticAnalysis(name={self.name})"
    
    def __repr__(self):
        return f"LinearStaticAnalysis(name={self.name})"


class LinearDynamicAnalysis:
    _option_fields = {
    }
    def __init__(self, name, job):
        self.name = name
        self._csi = job._tree
        self._job = job
        self.type = "LinDirHist"


    def run(self, asm):
        """
        Run the linear static analysis case.
        """
        self._apply_loads(asm)
        {
            "CASE - DIRECT HISTORY 3 - VISCOUS PROPORTIONAL DAMPING": [
                {"Case": "ACASE1", "SpecifyType": "Period", "MassCoeff": 0.0920869591121277, "StiffCoeff": 0.00433236240908684, "Period1": 1.4344, "Damping1": 0.02, "Period2": 1.29484, "Damping2": 0.02}
            ],
            "CASE - DIRECT HISTORY 4 - INTEGRATION PARAMETERS": [
                {"Case": "ACASE1", "IntMethod": "HilberHughesTaylor", "Gamma": 0.5, "Beta": 0.25, "Alpha": 0}
            ],
        }
        #
        # Damping
        #
        p03 = find_row(self._csi.get("CASE - DIRECT HISTORY 3 - VISCOUS PROPORTIONAL DAMPING", []), Case=self.name)
        if p03 is not None:
            model.rayleigh(p03.get("MassCoeff",0), 0, p03.get("StiffCoeff",0), 0)

        # Integration parameters
        beta = 0.25
        gamma = 0.5
        alpha = 0.0
        p04 = find_row(self._csi.get("CASE - DIRECT HISTORY 4 - INTEGRATION PARAMETERS", []), Case=self.name)
        if p04 is not None:
            beta = p04.get("Beta", beta)
            gamma = p04.get("Gamma", gamma)
            alpha = p04.get("Alpha", 0)

        # Run the analysis
        model = asm.model
        model.test("EnergyIncr", 1e-16, 10)
        model.system("Umfpack")
        model.constraints("Transformation")
        model.numberer("RCM")
        model.integrator("Newmark", beta=beta,gamma=gamma)
        model.algorithm("Newton")
        model.analysis("Transient")
        model.analyze(1)


    def _apply_loads(self, asm):
        """
        Apply loads to the model for this linear static analysis case.
        """

        for assign in find_rows(self._csi.get("CASE - DIRECT HISTORY 2 - LOAD ASSIGNMENTS", []),
                       Case=self.name):
            pattern = self._job.pattern(assign["LoadName"])
            pattern.apply(asm)

    def __str__(self):
        return f"LinearStaticAnalysis(name={self.name})"

    def __repr__(self):
        return f"LinearStaticAnalysis(name={self.name})"




class ModalAnalysis:
    """
    Uses tables:
    - "CASE - MODAL 1 - GENERAL"
    """
    _option_fields = {
        "solver",
        "max_num_modes",
        "min_num_modes",
    }
    def __init__(self, name, job, solver=None):
        self.name = name
        self._csi = job._tree
        self._job = job
        self.type = "LinModal"
        if solver is None:
            solver =  "-genBandArpack" #  "-fullGenLapack" # 
        self.solver = solver

        for assign in find_rows(self._csi.get("CASE - MODAL 1 - GENERAL", []),
                       Case=self.name):

            self.max_num_modes = assign.get("MaxNumModes",None)
            self.min_num_modes = assign.get("MinNumModes",1)

    @property 
    def options(self):
        return {
            "solver": self.solver,
            "max_num_modes": self.max_num_modes,
            "min_num_modes": self.min_num_modes,
        }

    def run(self, asm, verbose=True):
        """
        Run the modal analysis case.
        """
        self._apply_loads(asm)

        # Run the analysis
        model = asm.model
        model.constraints("Auto")
        ev = model.eigen(self.max_num_modes, solver=self.solver)
        # Print a summary of the last eigenvalue analysis 
        # (legacy OpenSeesPy command, will be changed in future)
        if verbose:
            model.modalProperties(print=True)
        return {
            "periods": [
                2 * 3.141592653589793 / (v ** 0.5) for v in ev
                if v > 1e-12
            ],
            "eigenvalues": ev
        }


    def _apply_loads(self, asm):
        """
        Apply loads to the model for this modal analysis case.
        """


    def __str__(self):
        return f"ModalAnalysis(name={self.name})"

    def __repr__(self):
        return f"ModalAnalysis(name={self.name})"


class NonlinearStaticAnalysis:
    """
    Uses tables:
    - "CASE - STATIC 1 - LOAD ASSIGNMENTS"
    - "CASE - STATIC 2 - NONLINEAR LOAD APPLICATION"
    - "CASE - STATIC 4 - NONLINEAR PARAMETERS"
    """
    def __init__(self, name, job):
        self.name = name
        self._csi = job._tree
        self._job = job
        csi = self._csi
        self.type = "NonStatic"

        self.SolScheme = "Iterative Only"

        p04 = find_row(csi.get("CASE - STATIC 4 - NONLINEAR PARAMETERS", []),
                       Case=name)

        MaxIterNR = 10
        if p04 is not None:
            self.SolScheme = p04.get("SolScheme", self.SolScheme)

            MaxIterNR = p04.get("MaxIterNR", MaxIterNR)
            ItConvTol = p04.get("ItConvTol", 1e-6)
            


    def _apply_loads(self, model):
        """
        Apply loads to the model for this nonlinear static analysis case.
        """

        "CASE - MODAL HISTORY 4 - PROPORTIONAL DAMPING"

        for assign in find_rows(self._csi.get("CASE - STATIC 1 - LOAD ASSIGNMENTS", []),
                       Case=self.name):
            pattern = self._job.pattern(assign["LoadName"])
            pattern.apply(model)

    def run(self, asm):
        model = asm.model
        self._apply_loads(model)

        if self.SolScheme != "Iterative Only":
            raise ValueError(f"Unsupported solution scheme: {self.SolScheme}")

        for assign in find_rows(self._csi.get("CASE - STATIC 2 - NONLINEAR LOAD APPLICATION", []),
                       Case=self.name):
            # pattern = self._job.pattern(assign["LoadName"])
            integrator = assign["LoadApp"]
        
        if integrator == "Displ Ctrl":
            u_max = assign["TargetDispl"]
            i_ctrl = assign["MonitorDOF"]

            # Use the Use Conjugate Displacement option if the analysis is having trouble converging.
            # The conjugate displacement is a weighted average of all displacements in the structure where each displacement degree of freedom is weighted by the load acting on that degree of freedom.
            # In other words, it is a measure of the work done by the applied load.
            model.integrator("DisplacementControl", 1, 1, 1.0, 1, 100)

        else:
            raise ValueError(f"Unsupported nonlinear load application: {integrator}")



#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#

from .utility import UnimplementedInstance, find_row, find_rows

def create_points(instance):
    assembly = instance.assembly
    names    = instance.names
    model    = instance.model
    csi      = instance._tree

    ndm = assembly.ndm
    ndf = assembly.ndf
#   dofs = config["dofs"]
    dofs = csi["ACTIVE DEGREES OF FREEDOM"][0]


    def _not_rotated(axes):
        return axes is None or (not axes.get("AdvanceAxes", False) and
          all(map(lambda i: axes[f"Angle{i}"]==0.0, "ABC")))

    for node in csi["JOINT COORDINATES"]:
        if find_row(csi.get("JOINT SPRING ASSIGNMENTS 2 - COUPLED",[]),Joint=node["Joint"]):
            names.log(UnimplementedInstance("Joint.Spring.Coupled", node))

        node_obj = assembly._make_node(node)

        node_tag = names.define("Joint", "node", node["Joint"])

        model.node(node_tag, node_obj.location)

        for i,v in enumerate(dofs.values()):
            if not v:
                model.fix(node_tag, dof=i+1)

        #
        # For any rotated nodes, create a duplicate node
        #
        axes = find_row(csi.get("JOINT LOCAL AXES ASSIGNMENTS 1 - TYPICAL", []),Joint=node["Joint"])

        springs = find_row(csi.get("JOINT SPRING ASSIGNMENTS 1 - UNCOUPLED", []),Joint=node["Joint"])

        if _not_rotated(axes):
            # if not rotated, the base node is just an alias for the real node
            if springs is None:
                node_base = names.define("JointBase", "node", node["Joint"], node_tag)
            else:
                # if springs are defined, create a separate base node
                spring_tag = names.define("JointBase", "node", node["Joint"])
                model.node(spring_tag, node_obj.location)
                for i,dof_key in enumerate(("U1","U2","U3","R1","R2","R3")):
                    if springs[dof_key] != 0.0:
                        mat = names.define("Material", "material")
                        model.uniaxialMaterial(
                            "Elastic",
                            mat,
                            springs[dof_key]
                        )
                        model.element(
                            "zeroLength",
                            names.define("Joint", "element"),
                            (spring_tag, node_tag),
                            mat=mat,
                            dir=i+1)
                    else:
                        model.equalDOF(node_tag, spring_tag, i+1)
                
        else:
            # if rotated, create aux node to represent the base;
            # no fourth argument means a new unique tag is created
            node_base = names.define("JointBase", "node", node["Joint"])
            assert node_base != node_tag
            assert names.identify("JointBase", "node", node["Joint"]) == node_base
            model.node(node_base, 
                       node_obj.location)
            # for i,v in enumerate(dofs.values()):
            #     if not v:
            #         model.fix(node_base, dof=i+1)
            model.constrain(node_base, 
                            node_tag, #node_base,
                            rotate=_joint_rotvec(instance, node, axes).tolist()
            )


                    
    for node in csi.get("JOINT RESTRAINT ASSIGNMENTS", []):
        node_tag = names.identify("JointBase", "node", node["Joint"])
        # Only fix dofs that werent aready fixed globally
        # Note that dof keys look like UX, RY, etc, but in the restraint
        # table they look like U1, R2, etc
        model.fix(node_tag, tuple(int(node[f"{key[0]}{'XYZ'.find(key[1])+1}"]) if dofs[key] else 0 for key in dofs))


    for node in csi.get("JOINT ADDED MASS ASSIGNMENTS", []):
        mass = [node[f"Mass{i+1}"] for i in range(ndm)]
        mass = mass + [0.0]*(ndf-len(mass))
        node_tag = names.identify("Joint", "node", node["Joint"])
        if node["CoordSys"].upper() != "GLOBAL":
            names.log(UnimplementedInstance(f"Joint.Mass.CoordSys={node['CoordSys']}", node))
        model.mass(node_tag, tuple(mass))


    for node in csi.get("JOINT ADDED MASS BY VOLUME ASSIGNMENTS", []):
        dens = find_row(csi.get("MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES",[]),
                        Material=node["Material"])["UnitMass"]
        vols = [node[f"Vol{i}"] for i in range(1,ndm+1) if f"Vol{i}" in node]
        vols = vols + [0.0]*(ndf-len(vols))
        mass = tuple(vol*dens for vol in vols)
        model.mass(node["Joint"], mass)


    "JOINT CONSTRAINT ASSIGNMENTS"
    "CONSTRAINT DEFINITIONS - BEAM"
    "CONSTRAINT DEFINITIONS - BODY"
    "CONSTRAINT DEFINITIONS - EQUAL"
    "CONSTRAINT DEFINITIONS - LINE"
    "CONSTRAINT DEFINITIONS - LOCAL"
    "CONSTRAINT DEFINITIONS - PLATE"
    "CONSTRAINT DEFINITIONS - ROD"
    "CONSTRAINT DEFINITIONS - WELD"
    "CONSTRAINT DEFINITIONS - BRIDGE OBJECT FLAGS"

    _apply_constraints(instance)

    _create_diaphragms(instance)


    # TODO: handle "JOINT MERGE NUMBER ASSIGNMENTS"

    return


def _joint_rotvec(instance, node, axes):
    names = instance.names
    csi = instance._tree

    if not axes.get("AdvanceAxes", False):
        from shps.rotor import log, exp
        from math import radians as _rad
        rotvec = log(
             exp([0.0, 0.0, _rad(axes["AngleA"])])
           @ exp([0.0, _rad(axes["AngleB"]), 0.0])
           @ exp([_rad(axes["AngleC"]), 0.0, 0.0])
        )
        return rotvec

    axes = find_row(csi.get("JOINT LOCAL AXES ASSIGNMENTS 2 - ADVANCED", []),Joint=node["Joint"])
    if axes is None:
        names.log(UnimplementedInstance(f"Joint.Restraint.AdvancedAxes=True without ADVANCED table"))

    return instance.assembly._coordinate_system(axes["AxCoordSys"]).rotation_vector


def _create_diaphragms(instance):
    tables = instance._tree
    model  = instance.model
    conv   = instance.names

    for dia in tables.get("CONSTRAINT DEFINITIONS - DIAPHRAGM", []):
        if dia.get("CoordSys", "GLOBAL").upper() != "GLOBAL":
            instance._log(UnimplementedInstance("Diaphragm.CoordSys!=GLOBAL", dia))
            continue
        if dia.get("Axis", None) not in ("X", "Y", "Z"):
            instance._log(UnimplementedInstance("Diaphragm.Axis not in (X,Y,Z)", dia))
            continue
        if dia.get("CoordSys", "GLOBAL").upper() != "GLOBAL":
            instance._log(UnimplementedInstance("Diaphragm.CoordSys!=GLOBAL", dia))
            continue
            

        joints = find_rows(tables.get("JOINT CONSTRAINT ASSIGNMENTS", []), Constraint=dia["Name"])
        if len(joints) < 2:
            instance._log(UnimplementedInstance("Diaphragm with less than 2 joints", dia))
            continue

        nodes = [conv.identify("JointBase", "node", joint["Joint"]) for joint in joints]

        model.rigidDiaphragm("XYZ".index(dia["Axis"])+1, tuple(nodes))


def _apply_constraints(ins):
    sap = ins._tree
    model = ins.model
    conv  = ins.names

    DOFS = ("UX","UY","UZ","RX","RY","RZ")
    # The format of body dictionary is {'node number':'constraint name'}
    for body in sap.get("CONSTRAINT DEFINITIONS - BODY", []):
        constraints = {}
        if not all(value for value in map(lambda x: body[x], DOFS)):
            ins._log(UnimplementedInstance("Joint.Constraint.Body.IncompleteDOF", body))
            continue

        for constraint in  find_rows(sap.get("JOINT CONSTRAINT ASSIGNMENTS",[]), Constraint=body["Name"]):
            if "Type" in constraint and constraint["Type"] == "Body":
                # map node number to constraint
                constraints[constraint["Joint"]] = constraint["Constraint"]
            else:
                ins._log(UnimplementedInstance("Joint.Constraint", constraint))

        # Sort the dictionary by body name and return a list [(node, body name)]
        constraints = list(sorted(constraints.items(), key=lambda x: x[1]))


        if len(constraints) > 0:
            nodes = []
            # Assign the first body name to the pointer
            pointer = constraints[0][1]

            # Traverse the tuple. If the second element in the tuple, the body
            # name, is the same as the pointer, then store the node number, 
            # into nodes.
            for node, constraint in constraints:
                if constraint == pointer:
                    nodes.append(node)
                else:
                    # First write the nodes in nodes to the body file
                    for le in range(len(nodes)-1):
                        ni = conv.identify("JointBase", "node", nodes[0])
                        nj = conv.identify("JointBase", "node", nodes[le + 1])
                        model.eval(f"rigidLink beam {ni} {nj}\n")
                    # Restore nodes and save the node that returns False.
                    nodes = []
                    nodes.append(node)
                    # The pointer is changed to the new body name
                    pointer = constraint

            # After the for loop ends, write the nodes in the nodes of the last loop to the body file.
            for le in range(len(nodes)-1):
                ni = conv.identify("JointBase", "node", nodes[0])
                nj = conv.identify("JointBase", "node", nodes[le + 1])
                model.eval(f"rigidLink beam {ni} {nj}\n")





from xcsi.csi.utility import find_row, find_rows, UnimplementedInstance
import numpy as np
import warnings
from xcsi.csi._material import find_material

from xsection import ElasticConstants
from xsection import PolygonSection as Polygon, CompositeSection
from xsection.library import (
    WideFlange, 
    Rectangle, 
    Equigon, 
    Circle, 
    HollowRectangle, 
    Angle, 
    Pipe
)

from xara.benchmarks import create_section as create_xara_section

CIRCLE_DIVS = 40

class FrameSection:
    def __init__(self, csi, prop_01, mode=None):
        self._mode = mode

        if isinstance(prop_01, str):
            name = prop_01
            prop_01 = find_row(csi.get("FRAME SECTION PROPERTIES 01 - GENERAL",[]), SectionName=name)
            if prop_01 is None:
                raise ValueError(f"Section {name} not found.")
        else:
            name = prop_01["SectionName"]
        
        self.name = name

        self._prop_01 = prop_01
        self._csi = csi

        self._elastic = None

    @property
    def has_shear(self):
        e = self.elastic()
        return e is not None and e.Ay != 0 and e.Az != 0
    

    @property 
    def has_uniaxial(self):
        material = self.material
        return False 
    
    @property 
    def has_multiaxial(self):
        material = self.material
        if material is not None:
            return material.get("E",None) is not None and \
                   material.get("G",None) is not None
        return False

    @property 
    def material(self):
        if "Material" not in self._prop_01:
            return None
        return find_material(self._csi, self._prop_01["Material"])

    @property 
    def mass(self):
        if "TotalMass" in self._prop_01:
            elements = find_rows(self._csi.get("FRAME SECTION ASSIGNMENTS", []), AnalSect=self.name)
            length = 0.0 
            for elem in elements:
                frame = find_row(self._csi.get("CONNECTIVITY - FRAME", []), Frame=elem["Frame"])
                length += frame["Length"]
            total =  self._prop_01["TotalMass"]/length
            return total
        A = self._prop_01.get("Area", 0.0)
        material = self.material
        rho = material["density"]
        return A * rho

    # Simulation
    def create_fibers(self, mesh_scale=None, **kwds):
        return self._create_model(mesh_scale=mesh_scale).create_fibers(**kwds)


    def elastic(self):
        prop_01 = self._prop_01
        csi = self._csi

        if self._elastic is None:
            self._elastic = _find_elastic(csi, prop_01)

        return self._elastic
    

    def _create_model(self, mesh_size=None, render_only=True):
        csi = self._csi
        prop_01 = self._prop_01
        name = self.name
        details = not render_only

        centroid = np.array([-prop_01.get("CGOffset3", 0), 
                              prop_01.get("CGOffset2", 0)])

        #
        # First skip over non-basic sections if we need detailed model
        #
        shape = self._prop_01.get("Shape", "SD Section")

        if details and (shape in {"SD Section", "Bridge Section"}):
            pass

        elif details and find_row(csi.get("FRAME SECTION PROPERTIES 02 - CONCRETE COLUMN",[]), SectionName=name):
            pass


        #
        # Begin matching homogeneous standard shapes
        #
        elif shape == "Circle":
            r = prop_01["t3"]/2
            return Circle(
                radius=r,
                divisions=CIRCLE_DIVS,
                z=0,
                mesh_scale=mesh_size,
                centroid=centroid,
            )

        elif shape == "I/Wide Flange":
            # Example 1-018a.s2k
            return WideFlange(
                d=prop_01["t3"],
                b=prop_01["t2"],
                tf=prop_01["tf"],
                tw=prop_01["tw"],
                centroid=centroid,
            )
        

        elif shape == "Rectangular":
            from xsection.library import Rectangle
            return Rectangle(d=prop_01["t3"], 
                             b=prop_01["t2"],
                             centroid=centroid,
            )
        
        # elif shape == "Pipe":
        #     return Pipe(radius=prop_01["t3"]/2,
        #                 thickness=prop_01["tw"],
        #                 centroid=centroid)
        
        elif shape=="Box/Tube":
            from xsection.library import HollowRectangle

            "SECTION DESIGNER PROPERTIES 09 - SHAPE BOX/TUBE"
            # t3=4.5   t2=4.5   tf=0.8   tw=0.8
            return HollowRectangle(
                d=prop_01["t3"],
                b=prop_01["t2"],
                tf=prop_01["tf"],
                tw=prop_01["tw"],
                centroid=centroid,
                # elastic = self.elastic()
            )
        

        elif shape == "SD Section":
            prop_sd = find_row(csi.get("SECTION DESIGNER PROPERTIES 01 - GENERAL", []), SectionName=name)
            shapes = []

            for shape in find_rows(csi.get("SD STRESS-STRAIN 06 - CONCRETE MANDER CONFINED CIRCLE", []), SectionName=name):

                if shape["Part"] == "Core": 
                    z = 1

                shapes.append(
                    Circle(
                        radius=shape["CnfDiam"]/2,
                        z=z)
                )
                
                bar = Circle(
                             radius=np.sqrt(shape["CnfBarArea"])/2, 
                             z=2, 
                             mesh_scale=1/2, 
                             divisions=4, 
                             name="rebar")

                shapes.extend([

                ])

            if prop_sd["nPolygon"] > 0:
                polygon_data = csi.get("SECTION DESIGNER PROPERTIES 16 - SHAPE POLYGON", [])
                # if mesh_size is None:
                #     raise ValueError("Mesh size must be provided for polygonal sections")
                interior = []
                exterior = None
                for srow in find_rows(polygon_data, SectionName=name):
                    if srow.get("ShapeMat","") != "Opening":
                        exterior =  np.array([
                            [row["X"], row["Y"]]
                            for row in (find_rows(polygon_data, SectionName=name, ShapeName=srow["ShapeName"]))  #row.get("ShapeName","")=="Polygon1"
                        ])
                        break

                if exterior is not None:

                    for hole in find_rows(polygon_data, SectionName=name, ShapeMat="Opening"):
                        interior.append(np.array([
                            [row["X"], row["Y"]]
                            for row in find_rows(polygon_data, ShapeName=hole["ShapeName"], SectionName=name)
                        ]))

                    shapes.append(
                        Polygon(exterior,
                                interior, 
                                mesh_size=1/100,
                        )
                    )

            for shape in find_rows(csi.get("SECTION DESIGNER PROPERTIES 07 - SHAPE ANGLE", []), SectionName=name):
                tf = shape["FlngThick"]
                tw = shape["WebThick"]
                assert tf == tw, "Angle section must have equal flange and web thicknesses"

                shapes.append(
                    Angle(
                        d=shape["Height"],
                        b=shape["Width"],
                        t=tf,
                        centroid=centroid,
                        z=shape.get("Zorder",0))
                )

            for circle in find_rows(csi.get("SECTION DESIGNER PROPERTIES 24 - SHAPE CALTRANS CIRCLE", []), SectionName=name):
                if details:
                    return None
    
                assert circle["Height"] == circle["Width"]
                r = circle["Height"]/2
                exterior = np.array([
                    [np.sin(x)*r, np.cos(x)*r] for x in np.linspace(0, np.pi*2, CIRCLE_DIVS)
                ])
                shapes.append(
                    Polygon(exterior, 
                               mesh_size=r/5,
                    )
                )


            for row in find_rows(csi.get("SECTION DESIGNER PROPERTIES 26 - SHAPE CALTRANS OCTAGON", []), SectionName=name):
                assert row["Height"] == row["Width"]
                shapes.append(
                    Equigon(
                        row["Height"]/2,
                        divisions=8,
                        z=0)
                )

                shapes.extend([

                ])

            if len(shapes) == 0:
                warnings.warn(f"Section {name} has no shapes defined in SECTION DESIGNER PROPERTIES 01 - GENERAL")
                return None
            elif len(shapes) == 1:
                return shapes[0]#.translate(-centroid)
            else:
                return CompositeSection(shapes)


        #
        # BRIDGE SECTIONS
        #
        elif shape == "Bridge Section":

            polygon_data = csi.get("FRAME SECTION PROPERTIES 06 - POLYGON DATA", [])

            cbg = find_row(csi.get("BRIDGE SECTION DEFINITIONS 02 - CONCRETE BOX GIRDER", []), Section=name)
            if cbg:
                pass

            elif polygon_data and not details:
                if mesh_size is None:
                    if render_only:
                        mesh_size=10
                    else:
                        raise ValueError("Mesh size must be provided for polygonal sections")
                
                interior = []
                exterior_row = find_row(polygon_data, SectionName=name, Opening=False)
                exterior =  np.array([
                    [row["X"], row["Y"]]
                    for row in sorted(find_rows(polygon_data, SectionName = name, Polygon=exterior_row["Polygon"]),
                                      key=lambda r: r["Point"], reverse=True)
                ])
                ref = (exterior_row["RefPtX"],  exterior_row["RefPtY"])

                for i in range(len(exterior)):
                    exterior[i] -= ref


                for hole in find_rows(polygon_data, SectionName=name, Opening=True):
                    interior.append(np.array([
                        [row["X"], row["Y"]]
                        for row in sorted(find_rows(polygon_data, Polygon=hole["Polygon"]), 
                                          key=lambda r: r["Point"], reverse=True)
                    ]))
                    for i in range(len(interior[-1])):
                        interior[-1][i] -= ref


                return Polygon(exterior,
                               interior, 
                               mesh_size=mesh_size,
                        )#.translate(-centroid)

#
#
#
def _find_elastic(csi, prop_01):

    material = find_material(csi, prop_01["Material"])
    E = material["E"]
    G = material["G"]
    Iy = prop_01["I33"]*prop_01.get("I3Mod", 1)
    Iz = prop_01["I22"]*prop_01.get("I2Mod", 1)
    if Iy == 0.0:
        Iy = Iz
    if Iz == 0.0:
        Iz = Iy
    J = prop_01["TorsConst"]*prop_01.get("JMod", 1)
    if J == 0.0:
        J = (Iy + Iz)*100

    return ElasticConstants(
            A  = prop_01["Area"]*prop_01.get("AMod", 1),
            Ay = prop_01["AS3"]*prop_01.get("A3Mod", 1),
            Az = prop_01["AS2"]*prop_01.get("A2Mod", 1),
            Iy = Iy,
            Iz = Iz,
            J  = J,
            E  = E,
            G  = G
        )
    return ElasticConstants(
            A  = prop_01["Area"]*prop_01.get("AMod", 1),
            Ay = prop_01["AS3"]*prop_01.get("A2Mod", 1),
            Az = prop_01["AS2"]*prop_01.get("A3Mod", 1), # TODO: check this!!!
            Iy = prop_01["I22"]*prop_01.get("I2Mod", 1),
            Iz = prop_01["I33"]*prop_01.get("I3Mod", 1),
            J  = prop_01["TorsConst"]*prop_01.get("JMod", 1),
            E  = E,
            G  = G
        )



import json 

class DepthLimitedEncoder(json.JSONEncoder):
    def __init__(self, *args, max_depth=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_depth = max_depth

    def encode(self, o, _depth=0):
        if isinstance(o, dict):
            if _depth >= self.max_depth:
                # compact formatting
                return "{" + ", ".join(
                    f"{json.dumps(k)}: {self.encode(v, _depth+1)}"
                    for k, v in o.items()
                ) + "}"
            else:
                # pretty formatting with newlines/indents
                items = []
                for k, v in o.items():
                    items.append(
                        self.indent_str(_depth+1) +
                        f"{json.dumps(k)}: {self.encode(v, _depth+1)}"
                    )
                return "{\n" + ",\n".join(items) + "\n" + self.indent_str(_depth) + "}"

        elif isinstance(o, list):
            if _depth >= self.max_depth:
                return "[" + ", ".join(self.encode(v, _depth+1) for v in o) + "]"
            else:
                items = []
                for v in o:
                    items.append(
                        self.indent_str(_depth+1) + self.encode(v, _depth+1)
                    )
                return "[\n" + ",\n".join(items) + "\n" + self.indent_str(_depth) + "]"

        else:
            return json.dumps(o)

    def indent_str(self, depth):
        return " " * (self.indent * depth if self.indent else 0)


#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
"""
"SECTION DESIGNER PROPERTIES 01 - GENERAL"
"SECTION DESIGNER PROPERTIES 02 - REINFORCING AT SHAPE EDGES"
"SECTION DESIGNER PROPERTIES 03 - REINFORCING AT SHAPE CORNERS"
"SECTION DESIGNER PROPERTIES 04 - SHAPE I/WIDE FLANGE"
"SECTION DESIGNER PROPERTIES 05 - SHAPE CHANNEL"
"SECTION DESIGNER PROPERTIES 06 - SHAPE TEE"
"SECTION DESIGNER PROPERTIES 07 - SHAPE ANGLE"
"SECTION DESIGNER PROPERTIES 08 - SHAPE DOUBLE ANGLE"
"SECTION DESIGNER PROPERTIES 09 - SHAPE BOX/TUBE"
"SECTION DESIGNER PROPERTIES 10 - SHAPE PIPE"
"SECTION DESIGNER PROPERTIES 11 - SHAPE PLATE"
"SECTION DESIGNER PROPERTIES 12 - SHAPE SOLID RECTANGLE"
"SECTION DESIGNER PROPERTIES 13 - SHAPE SOLID CIRCLE"
"SECTION DESIGNER PROPERTIES 14 - SHAPE SOLID SEGMENT"
"SECTION DESIGNER PROPERTIES 15 - SHAPE SOLID SECTOR"
"SECTION DESIGNER PROPERTIES 16 - SHAPE POLYGON"

"SECTION DESIGNER PROPERTIES 17 - SHAPE REINFORCING SINGLE"
"SECTION DESIGNER PROPERTIES 18 - SHAPE REINFORCING LINE"
"SECTION DESIGNER PROPERTIES 19 - SHAPE REINFORCING RECTANGLE"
"SECTION DESIGNER PROPERTIES 20 - SHAPE REINFORCING CIRCLE"

"SECTION DESIGNER PROPERTIES 21 - SHAPE REFERENCE LINE"
"SECTION DESIGNER PROPERTIES 22 - SHAPE REFERENCE CIRCLE"

"SECTION DESIGNER PROPERTIES 23 - SHAPE CALTRANS SQUARE"
"SECTION DESIGNER PROPERTIES 24 - SHAPE CALTRANS CIRCLE"
"SECTION DESIGNER PROPERTIES 25 - SHAPE CALTRANS HEXAGON"
"SECTION DESIGNER PROPERTIES 26 - SHAPE CALTRANS OCTAGON"

"SECTION DESIGNER PROPERTIES 27 - CALTRANS LONGITUDINAL REBAR"
"SECTION DESIGNER PROPERTIES 28 - CALTRANS LONGITUDINAL TENDONS"
"SECTION DESIGNER PROPERTIES 29 - CALTRANS CONFINEMENT REBAR"
"SECTION DESIGNER PROPERTIES 30 - FIBER GENERAL"
"SECTION DESIGNER PROPERTIES 31 - STRESS POINT"
"SECTION DESIGNER PROPERTIES 32 - BRIDGE SECTION SHELL LAYOUT"
"SECTION DESIGNER PROPERTIES 33 - BRIDGE SECTION SOLID LAYOUT"
"SECTION DESIGNER PROPERTIES 34 - BRIDGE SECTION CUTLINE"
"SECTION DESIGNER PROPERTIES 35 - BRIDGE SECTION CENTERLINE"
"""
from ..utility import find_row, find_rows, UnimplementedInstance
import numpy as np
from ...names import Names
from xcsi.csi._material import find_material
from xcsi._assembly.frame import FrameSection

from xsection import ElasticConstants
from xsection import PolygonSection as Polygon, CompositeSection
from xsection.library import WideFlange, Rectangle, Equigon, Circle, HollowRectangle, Angle, Pipe

from xara.benchmarks import create_section as create_xara_section

class FrameSectionPolicy:
    pass


def add_frame_sections(csi, model, conv, policy=None):

    if len(csi.get("CONNECTIVITY - FRAME", [])) == 0:
        return 

    for sect in csi.get("FRAME SECTION PROPERTIES 01 - GENERAL", []):

        # Ensure this section has not already been defined in the xara.Model
        if not conv.identify("AnalSect", "section",     sect["SectionName"]) and \
           not conv.identify("AnalSect", "integration", sect["SectionName"]):

            if (s:= create_section(csi, sect, conv)) is not None:
                try:
                    shape = s._create_model()
                except:
                    shape = None

                has_material = "Material" in sect and sect["Material"] is not None and sect["Material"] != ""

                # 1) Multiaxial
                if False: #s.has_multiaxial and has_material and shape is not None:
                    tag  = conv.define("AnalSect", "section", f"Multiaxial/{s.name}")
                    mtag = conv.identify("Material", "material", sect["Material"])
                    create_xara_section(model, "ShearFiber", tag, shape, mtag, shear=False)

                # 2) Uniaxial
                if s.has_uniaxial and has_material and shape is not None:
                    # TODO: define uniaxial section
                    pass

                # 3) Elastic
                if (e := s.elastic()) is not None:
                    model.section("FrameElastic",
                            conv.define("AnalSect", "section", sect["SectionName"]), #self.index,
                            A  = e.A,
                            Ay = e.Ay,
                            Az = e.Az,
                            Iz = e.Iz,
                            Iy = e.Iy,
                            J  = e.J,
                            E  = e.E,
                            G  = e.G
                    )
                continue

            if (sections := _create_integration(csi, sect)) is not None:
                tags = []
                for s in sections:
                    tag = conv.define("AnalSect", "section", sect["SectionName"])
                    tags.append(tag)

                    if (e := s[0].elastic()) is not None:
                        model.section("FrameElastic",
                                    tag,
                                    A  = e.A,
                                    Ay = e.Ay,
                                    Az = e.Az,
                                    Iz = e.Iz,
                                    Iy = e.Iy,
                                    J  = e.J,
                                    E  = e.E,
                                    G  = e.G
                        )

                model.beamIntegration("UserDefined",
                          conv.define("AnalSect", "integration", sect["SectionName"]),
                          len(sections),
                          tuple(tags),
                          tuple(i[1] for i in sections),
                          tuple(i[2] for i in sections))
                continue

            conv.log(UnimplementedInstance(f"FrameSection.Shape={sect['Shape']}"))
            # assert False, sect

    return


def iter_sections(csi, names=None, table=None):
    if names is None:
        names = Names()

    if table is None:
        table = "FRAME SECTION PROPERTIES 01 - GENERAL"

    for sect in csi.get(table, []):

        if not names.identify("AnalSect", "section",     sect["SectionName"]) and \
           not names.identify("AnalSect", "integration", sect["SectionName"]):
            if (s:= create_section(csi, sect, names)) is not None:
                yield s, sect["SectionName"]

            else:
                names.log(UnimplementedInstance(f"FrameSection.Shape={sect['Shape']}"))
                continue



def create_section(csi, prop_01, names=None,
                    elastic_only=False,
                    render_only=False) ->"FrameSection":
    # NOTE: This function is called from IRiE

    if names is None:
        names = Names()

    #
    if isinstance(prop_01, str):
        name = prop_01
        prop_01 = find_row(csi.get("FRAME SECTION PROPERTIES 01 - GENERAL",[]), SectionName=name)
        if prop_01 is None:
            prop_01 = find_row(csi.get("FRAME SECTION PROPERTIES - BRIDGE OBJECT FLAGS",[]), SectionName=name)
            if prop_01 is None:
                raise ValueError(f"Section {name} not found in either table.")
    else:
        name = prop_01["SectionName"]

    #
    # 1)
    #
    segments = find_rows(csi.get("FRAME SECTION PROPERTIES 05 - NONPRISMATIC",[]),
                            SectionName=name)


    if prop_01.get("Shape",None) not in {"Nonprismatic"}:
        s = FrameSection(csi, prop_01)
        return s

    # Note: if len(segments) == 1 it will be handled in _create_integration
    # 2)
    if prop_01["Shape"] == "Nonprismatic" and len(segments) != 1 and all(segment["StartSect"] == segment["EndSect"] for segment in segments): #section["NPSecType"] == "Advanced":

        # TODO: Currently just treating advanced as normal prismatic section

        if not names.identify("AnalSect", "section", prop_01["SectionName"]) : #segments[0]["StartSect"]):


            # find properties
            p = find_row(csi["FRAME SECTION PROPERTIES 01 - GENERAL"],
                                SectionName=segments[0]["StartSect"])

            assert p is not None

            return FrameSection(csi, p)


def _create_integration(csi, prop_01):
    # 3)
    segments = find_rows(csi["FRAME SECTION PROPERTIES 05 - NONPRISMATIC"],
                            SectionName=prop_01["SectionName"])

    if prop_01["Shape"] != "Nonprismatic" or len(segments) != 1: 
        return None


    # Interpolate linear-elastic sections properties
    assert len(segments) == 1

    segment = segments[0]

    # Create property interpolation
    def interpolate(point, prop):
        si = find_row(csi["FRAME SECTION PROPERTIES 01 - GENERAL"],
                            SectionName=segment["StartSect"]
        )
        sj = find_row(csi["FRAME SECTION PROPERTIES 01 - GENERAL"],
                            SectionName=segment["EndSect"]
        )

        # TODO: Taking material from first section assumes si and sj have the same
        # material
        material = find_row(csi["MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES"],
                            Material=si["Material"]
        )
        assert material == find_row(csi["MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES"],
                            Material=sj["Material"]
        )

        if prop in material:
            start = end = material[prop]
        else:
            start = si[prop]
            end = sj[prop]

        power = {
            "Linear":    1,
            "Parabolic": 2,
            "Cubic":     3
        }[segment.get(f"E{prop}Var", "Linear")]

        return start*(1 + point*((end/start)**(1/power)-1))**power

    #
    # Define a numerical integration scheme
    #
    from numpy.polynomial.legendre import leggauss
    nip = 5
    sections = []
    for x,wi in zip(*leggauss(nip)):
        xi = (1+x)/2

        sec = _InterpolatedSection(
            ElasticConstants(
                        A  = interpolate(xi, "Area"),
                        Ay = interpolate(xi, "AS2"),
                        Az = interpolate(xi, "AS2"),
                        Iz = interpolate(xi, "I33"),
                        Iy = interpolate(xi, "I22"),
                        J  = interpolate(xi, "TorsConst"),
                        E  = interpolate(xi, "E1"),
                        G  = interpolate(xi, "G12")
            )
        )

        sections.append((sec, xi, wi/2))


    return sections



class _InterpolatedSection: #(FrameSection):
    def __init__(self, elastic):
        self._elastic = elastic
        self._plastic = None

    def elastic(self):
        return self._elastic

    def plastic(self):
        return self._plastic




from xcsi.csi.utility import find_row, find_rows, UnimplementedInstance

def find_material(csi, material_name):
    E = None 
    G = None 

    material = find_row(csi.get("MATERIAL PROPERTIES 01 - GENERAL", []),
                        Material=material_name)

    if material is not None and "E" in material:
        E = material.get("E", None)
        if "G" in material:
            G = material["G"]
        elif "U" in material:
            G = E/2/(1+material.get("U", None))
    
    if E is None or G is None:
        material = find_row(csi.get("MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES", []),
                            Material=material_name
        )
        if material is not None:
            E = material.get("E1", None)
            if "G12" in material:
                G = material["G12"]
            elif "U12" in material:
                G = E/2/(1+material["U12"])

    return {
        "E": E,
        "G": G,
        "v": ( None if E is None or G is None else E/(2*G) - 1),
        "density": (material.get("UnitMass", 0) if material is not None else 0),
    }


def create_materials(instance):
    "SD STRESS-STRAIN 01 - REBAR PARK"
    "SD STRESS-STRAIN 02 - REBAR SIMPLE"
    "SD STRESS-STRAIN 03 - STRUCTURAL STEEL"
    "SD STRESS-STRAIN 04 - CONCRETE SIMPLE"
    "SD STRESS-STRAIN 05 - CONCRETE MANDER UNCONFINED"
    "SD STRESS-STRAIN 06 - CONCRETE MANDER CONFINED CIRCLE"
    "SD STRESS-STRAIN 07 - CONCRETE MANDER CONFINED RECTANGLE"
    conv = instance.names
    model = instance.model
    csi = instance._tree

    #
    # 1) Material
    #

    for mat in csi.get("MATERIAL PROPERTIES 01 - GENERAL", []):
        E = None
        poisson = None
        density = None
        Fc = None
        Fy = None
        Fu = None
        Hi = None
        name = mat["Material"]

        if (p02 := find_row(csi.get("MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES", []), Material=mat["Material"])):
            E = p02["E1"]
            poisson = p02.get("U12", None)
            density = p02.get("UnitMass", None)
        
        if (p03a := find_row(csi.get("MATERIAL PROPERTIES 03A - STEEL DATA", []), Material=mat["Material"])):
            Fy = p03a.get("EffFy", p03a.get("Fy", None))
            Fu = p03a.get("EffFu", p03a.get("Fu", None))
            if "SHard" in p03a:
                Hi = p03a["SHard"]*E 

        if (p03j := find_row(csi.get("MATERIAL PROPERTIES 03J - COUPLED NONLINEAR VON MISES DATA", []), Material=mat["Material"])):
            Fy = p03j.get("YieldStress", None)


        if mat.get("SymType", "") == "Isotropic" and E is not None and Fy is None:
            model.nDMaterial(
                "ElasticIsotropic",
                conv.define("Material", "material", name),
                E,       # Young
                poisson, # Poisson
            )

        elif mat.get("SymType", "")  == "Uniaxial":
            # TODO: implement other uniaxial materials
            pass

        elif mat.get("SymType", "")  == "Isotropic" and E is not None and Fy is not None:
            # model.nDMaterial(
            #     "ElasticIsotropic",
            #     conv.define("Material", "material", name),
            #     E,       # Young
            #     poisson, # Poisson
            # )

            model.nDMaterial(
                "J2",
                conv.define("Material", "material", name),
                E=E,  # Young
                nu=poisson, # Poisson
                Fy=Fy, # Yield
                Hiso=Hi or 0,
                Fs=Fy,
                Hsat=0
            )

        else:
            conv.log(UnimplementedInstance(f"Material.SymType={mat.get('SymType', '')}", mat))
            continue
        

